package jelite.engine.galaxy;

import java.io.*;

import jelite.*;
import liaquay.math.vector.*;
import liaquay.collection.*;

public class Sector extends ArrayList
{
  public static final float SIZE = Parameters.SIZE_OF_SECTOR;

  private boolean _checkFlags[][][] = new boolean[ 3 ][ 3 ][ 3 ];
  private Sector _neighbours[][][] = new Sector[ 3 ][ 3 ][ 3 ];
  private SolarSystem _solarSystem;
  private int _x;
  private int _y;
  private int _z;

  public final int getX()
  {
    return _x;
  }

  public final int getY()
  {
    return _y;
  }

  public final int getZ()
  {
    return _z;
  }

  public final boolean isSector( int x, int y, int z )
  {
    return ( _x == x ) && ( _y == y ) && ( _z == z );
  }

  public Sector( SolarSystem solarSystem, int x, int y, int z )
  {
    System.out.println( "New sector:" + x + " " + y + " " + z );

    _solarSystem = solarSystem;
    _x = x;
    _y = y;
    _z = z;

    cleanCheckFlags();
    cleanNeighbours();
  }

  public final SolarSystem getSolarSystem()
  {
    return _solarSystem;
  }

  public final void remove( Item item )
  {
    System.out.println( "Removing " + item );
    super.remove( item );

    // Inform the solar system that the item has been removed
    _solarSystem.itemRemoved( item );

    // So the item may be replaced in the future.
    item.setRemove( false );

    if( item.getDeceased() )
    {
      // Tidy up anthing ...

    }
  }

  public final void addNewItem( Item item )
  {
    add( item );
    item.setSector( this );
    _solarSystem.itemAdded( item );
  }

  public final void addNeighbour( Sector neighbour, int x, int y, int z )
  {
    _neighbours[ x ][ y ][ z ] = neighbour;
  }

  public final void cleanCheckFlags()
  {
    for( int i = 0; i < 3; ++ i )
    {
      for( int j = 0; j < 3; ++j )
      {
        for( int k = 0; k < 3; ++k )
        {
          _checkFlags[ i ][ j ][ k ] = false;
        }
      }
    }
  }

  public final void cleanNeighbours()
  {
    for( int i = 0; i < 3; ++ i )
    {
      for( int j = 0; j < 3; ++j )
      {
        for( int k = 0; k < 3; ++k )
        {
          _neighbours[ i ][ j ][ k ] = null;
        }
      }
    }
  }

  public void interact( long ms, long period )
  {
    for( int i = 0; i < 3; ++ i )
    {
      for( int j = 0; j < 3; ++j )
      {
        for( int k = 0; k < 3; ++k )
        {
          if( !_checkFlags[ i ][ j ][ k ] )
          {
            if( ( i == 1 ) && ( j == 1 ) && ( k == 1 ) )
            {
              interactLocal( ms, period );
            }
            else
            {
              Sector neighbour = _neighbours[ i ][ j ][ k ];

              if( neighbour != null )
              {
                for( int l = 0; l < getNumberOfItems(); ++l )
                {
                  Item item = getItem( l );

                  neighbour.interact( ms, period, item );
                }

                neighbour._checkFlags[ 2 - i ][ 2 - j ][ 2 - k ] = true;
              }

              _checkFlags[ i ][ j ][ k ] = true;
            }
          }
        }
      }
    }
  }

  private void checkCollision( long ms, long period, Item item1, Item item2 )
  {
    if( !item1.getRemove() && !item2.getRemove() )
    {
      if( item1.checkCollision( item2 ) )
      {
        Collision collision = item1.getCollision( item2 );

        if( collision == null )
        {
          collision = new Collision( ms );

          item1.addCollision( item2, collision );
          item2.addCollision( item1, collision );

          item1.collision( item2, 0 );
          item2.collision( item1, 0 );
        }
        else
        {
          long update = collision.getUpdate();

          collision.setUpdate( ms );

          if( update != ms )
          {
            item1.collision( item2, period );
            item2.collision( item1, period );
          }
        }
      }
    }
  }

  private void interact(  long ms, long period, Item item1 )
  {
    for( int i = 0; i < getNumberOfItems(); ++i )
    {
      Item item2 = getItem( i );

      checkCollision( ms, period, item1, item2 );
    }
  }

  private void interactLocal( long ms, long period )
  {
    for( int i = 0; i < getNumberOfItems(); ++ i )
    {
      for( int j = i + 1; j < getNumberOfItems(); ++j )
      {
        Item item1 = getItem( i );
        Item item2 = getItem( j );

        checkCollision( ms, period, item1, item2 );
      }
    }

    ArrayList oldCollisions = null;

    for( int i = 0; i < getNumberOfItems(); ++ i )
    {
      Item item1 = getItem( i );

      if( oldCollisions != null )
      {
        oldCollisions.removeAll();
      }

      if( item1.getNumberOfCollisions() > 0 )
      {
        for( Hashtable.HashtableValueIterator it = item1.getCollisionsIterator(); it.hasNext(); )
        {
          Item item2 = (Item)it.getKey();
          Collision collision = (Collision)it.next();

          if( collision.getUpdate() != ms )
          {
            if( oldCollisions == null )
            {
              oldCollisions = new ArrayList( 10 );
            }

            oldCollisions.add( item2 );
          }
        }

        if( oldCollisions != null )
        {
          for( int j = 0; j < oldCollisions.size(); ++j )
          {
            Item item2 = (Item)oldCollisions.get( j );

            item1.removeCollision( item2 );
            item2.removeCollision( item1 );
          }
        }
      }
    }
  }

  public final void removeFromNeighbours()
  {
    for( int i = 0; i < 3; ++ i )
    {
      for( int j = 0; j < 3; ++j )
      {
        for( int k = 0; k < 3; ++k )
        {
          Sector neighbour = _neighbours[ i ][ j ][ k ];

          if( neighbour != null )
          {
            neighbour._neighbours[ 2 - i ][ 2 - j ][ 2 - k ] = null;
          }
        }
      }
    }
  }

  /**
   * Perform item movement.
   */
  public final void update( long ms, long period )
  {
    cleanCheckFlags();
    for( int i = 0; i < getNumberOfItems(); )
    {
      Item item = getItem( i );

      if( item.getRemove() )
      {
        remove( item );
      }
      else
      {
        item.update( ms, period, null );
        ++i;
      }
    }
  }

  /**
   * This should be called after all items have been updated.
   * It will move items to their new homes.
   */
  public final void moveItems()
  {
    cleanCheckFlags();

    for( int i = 0; i < getNumberOfItems(); ++i )
    {
      Item item = getItem( i );

      // Fetch the sector relative sector coordinates.
      // These will be (0,0,0) if the item is still in the sector.
      int x = item.getSectorX();
      int y = item.getSectorY();
      int z = item.getSectorZ();

      // Check if the item has left the sector
      if( ( x != 0 ) || ( y != 0 ) || ( z != 0 ) )
      {
        System.out.println( "Leaving sector:" + _x + " " + _y + " " + _z );

        // Remove the item from this sector
        // This will NOT inform the solar system of the removal.
        // Modify the loop counter so that all items are checked.
        super.remove( item );
        --i;

        // Item coordinates are sector relative so they need adjusting.
        item.trimToSector();

        // Attempt to find the new sector in neighbours as this will
        // save on some lookups.
        Sector newSector = null;
        if( ( x >= -1 ) && ( x <= 1 ) &&
            ( y >= -1 ) && ( y <= 1 ) &&
            ( z >= -1 ) && ( z <= 1 ) )
        {
          newSector = _neighbours[ x + 1 ][ y + 1 ][ z + 1 ];
        }

        if( newSector == null )
        {
          // Convert from relative to absolute sector coordinates
          int newSectorX = _x + x;
          int newSectorY = _y + y;
          int newSectorZ = _z + z;

          newSector = _solarSystem.getGalaxy().getSector( _solarSystem, newSectorX, newSectorY, newSectorZ );
        }

        // As the sector exits extract relavent information
        SolarSystem newSolarSystem = newSector.getSolarSystem();

        // Check if the neighbour sector is in another solar system
        if( _solarSystem == newSolarSystem )
        {
          // The new sector is in this solar system
          // Add the item to the new sector.
          // The item has not moved solar system so nobody needs be informed.
          newSector.add( item );
          item.setSector( newSector );
        System.out.println( "Itra-solar-system move" );
        }
        else
        {
          // The new sector is in another solar system.
          // Add the item to the new sector.
          newSector.addNewItem( item );

          // The item has moved solar system so inform the solar systems.
          _solarSystem.itemRemoved( item );
        System.out.println( "Iter-solar-system move" );
        }

        System.out.println( "Item " + item + " added to " + newSector );
        System.out.println( "Item sector " + item.getSector().getX() + " " + item.getSector().getY() + " " +item.getSector().getZ() );
      }
    }
  }

  public String toString()
  {
    String s = "Sector " + _x + " " + _y + " " + _z;
    s = s + "\nNumber of items "  + size();
    for( int i = 0; i < 3; ++ i )
    {
      for( int j = 0; j < 3; ++j )
      {
        for( int k = 0; k < 3; ++k )
        {
          Sector neighbour = _neighbours[ i ][ j ][ k ];

          if( neighbour != null )
          {
            s = s + "\nNeighbour"  + neighbour._x + " " + neighbour._y + " " + neighbour._z;
          }
        }
      }
    }
    return s;
  }

  public final Item getItem( int i )
  {
    return (Item)get( i );
  }

  public final int getNumberOfItems()
  {
    return length();
  }

  public Sector[][][] getNeighbours()
  {
    return _neighbours;
  }

  public void addSound( int soundIdentifier, long period, boolean loop )
  {
    for( int i = 0; i < getNumberOfItems(); ++i )
    {
      Item item = getItem( i );
      item.addSound( soundIdentifier, period, loop );
    }
  }

  public void addSoundNeighbours( int soundIdentifier, long period, boolean loop )
  {
    addSound( soundIdentifier, period, loop );
    for( int i = 0; i < 3; ++ i )
    {
      for( int j = 0; j < 3; ++j )
      {
        for( int k = 0; k < 3; ++k )
        {
          Sector neighbour = _neighbours[ i ][ j ][ k ];

          if( neighbour != null )
          {
            neighbour.addSound( soundIdentifier, period, loop );
          }
        }
      }
    }
  }
}