package jelite.engine.galaxy;

import java.io.*;
import java.awt.*;

import jelite.*;
import liaquay.collection.*;
import liaquay.math.vector.*;

public class SolarSystem extends ArrayList
{
  public static final int SIZE_IN_SECTORS = Parameters.SOLAR_SYSTEM_SIZE_IN_SECTORS;
  private static final int HALF_SIZE_IN_SECTORS = ( SIZE_IN_SECTORS - 1 ) / 2;
  private Galaxy _galaxy;
  private int _x;
  private int _y;
  private int _z;
  private int _itemCount = 0;
  private Star _star = null;
  private ArrayList _satellites = new ArrayList();

  public SolarSystem( Galaxy galaxy, int x, int y, int z, Star star )
  {
    System.out.println( "New solar system:" + x + " " + y + " " + z );

    _galaxy = galaxy;
    _x = x;
    _y = y;
    _z = z;
    _star = star;

    if( _star != null )
    {
      Sector origin = getSector( _x * SIZE_IN_SECTORS, _y * SIZE_IN_SECTORS, _z * SIZE_IN_SECTORS );
      Item starItem = new Item( _star.getRenderable(), new Transform() );
      starItem.setScenery( true );
      origin.addNewItem( starItem );
      addSatellites( origin, starItem, _star.getSatellites() );
    }
  }

  public Star getStar()
  {
    return _star;
  }

  public ArrayList getSatellites()
  {
    return _satellites;
  }

  private final void addSatellites( Sector origin, Item starItem, ArrayList satellites )
  {
    for( int i = 0; i < satellites.size(); ++i )
    {
      Satellite satellite = (Satellite)satellites.get( i );
      SatelliteItem satelliteItem;
      if( satellite instanceof SpaceStationSatellite )
      {
        //satelliteItem = new SpaceStationItem( satellite, starItem );
         satelliteItem = new SatelliteItem( satellite, starItem );
      }
      else
      {
        satelliteItem = new SatelliteItem( satellite, starItem );
      }
      satelliteItem.setScenery( true );
      origin.addNewItem( satelliteItem );
      _satellites.add( satelliteItem );

      addSatellites( origin, starItem, satellite.getSatellites() );
    }
  }

  public final boolean isSolarSystem( int x, int y, int z )
  {
    return ( _x == x ) && ( _y == y ) && ( _z == z );
  }

  public final String getName()
  {
    if( _star != null )
    {
      return _star.getName();
    }
    else
    {
      return "Unknown";
    }
  }

  /**
   * Convert a sector coordinate to a galaxy coordinate.
   * Must be an odd number of sectors to galaxy.
   */
  public static final int sectorToGalaxyCoordinate( int sectorCoordinate )
  {
    if( sectorCoordinate > 0 )
    {
      return ( ( sectorCoordinate + HALF_SIZE_IN_SECTORS ) / SIZE_IN_SECTORS );
    }
    else
    {
      return ( ( sectorCoordinate - HALF_SIZE_IN_SECTORS ) / SIZE_IN_SECTORS );
    }
  }

  public SolarSystem( Galaxy galaxy )
  {
    _galaxy = galaxy;
  }

  public final Galaxy getGalaxy()
  {
    return _galaxy;
  }

  public final int getX()
  {
    return _x;
  }

  public final int getY()
  {
    return _y;
  }

  public final int getZ()
  {
    return _z;
  }

  public final Sector findSector( int x, int y, int z )
  {
    for( int i = 0; i < length(); ++i )
    {
      Sector s = (Sector)get( i );
      if( s.isSector( x, y, z ) )
      {
        return s;
      }
    }
    return null;
  }

  public final Sector createSector( int x, int y, int z )
  {
    Sector s = new Sector( this, x, y, z );

    for( int i = -1; i < 2; ++i )
    {
      for( int j = -1; j < 2; ++j )
      {
        for( int k = -1; k < 2; ++k )
        {
          if( ( i != 0 ) || ( j != 0 ) || ( k != 0 ) )
          {
            Sector neighbour = _galaxy.findSector( this, x + i, y + j, z + k );

            if( neighbour !=  null )
            {
              s.addNeighbour( neighbour, i + 1, j + 1, k + 1 );
              neighbour.addNeighbour( s, 1 - i, 1 - j, 1 - k );
            }
          }
        }
      }
    }

    add( s );

    return s;
  }

  public final Sector getSector( int x, int y, int z )
  {
    Sector s = findSector( x, y, z );

    if( s == null )
    {
      s = createSector( x, y, z );
    }

    return s;
  }

  public final void add( Item item, int x, int y, int z )
  {
    System.out.println( "Adding item " + item + " to " + x + " " + y + " " + z );
    Sector s = getSector( x, y, z );

    s.add( item );

    item.setSector( s );
  }

  public final void add( Item item, Sector s )
  {
    System.out.println( "Adding item " + item + " to " + s );

    s.add( item );

    item.setSector( s );
  }

  public final void add( Item item )
  {
    int x = item.getSectorX();
    int y = item.getSectorY();
    int z = item.getSectorZ();

    Sector s = getSector( x, y, z );

    item.trimToSector();

    s.add( item );

    item.setSector( s );
  }

  public final void itemRemoved( Item item )
  {
    if( !item.getScenery() )
    {
      --_itemCount;
    }
    System.out.println( "Item removed from " + this + ". Now contains " + _itemCount + " items."  );
  }

  public final void itemAdded( Item item )
  {
    if( !item.getScenery() )
    {
      ++_itemCount;
    }
    System.out.println( "Item added to " + this + ". Now contains " + _itemCount + " items." );
  }

  public final int getItemCount()
  {
    return _itemCount;
  }

  private final void removeSector( Sector s )
  {
    s.removeFromNeighbours();

    remove( s );
  }

  public final void interact( long ms, long period )
  {
    for( int i = 0; i < length(); ++i )
    {
      Sector s = (Sector)get( i );

      s.interact( ms, period );

      // Remove empty sectors
      if( s.getNumberOfItems() == 0 )
      {
        removeSector( s );
        --i;
      }
    }
  }

  public final void update( long ms, long period )
  {
    if( _star != null )
    {
      _star.calculatePosition( ms );
    }

    for( int i = 0; i < length(); ++i )
    {
      Sector s = (Sector)get( i );
      s.update( ms, period );
    }
  }

  public final void moveItems()
  {
    for( int i = 0; i < length(); ++i )
    {
      Sector s = (Sector)get( i );
      s.moveItems();
    }
  }
}