package liaquay.collection;



/**
 * Manages a collection of objects as an array in memory.
 * @author Philip M. Scull
 */
public class ArrayList implements ObjectArray
{
  private Object[] _objects;
  private int _length = 0;

  /**
   * Creates an array-list with an initial capacity of 50 objects.
   */
  public ArrayList()
  {
    _objects = new Object[ 50 ];
  }

  /**
   * Creates an array-list with a specified initial capacity.
   * @param initialLength The initial capacity of the array-list.
   */
  public ArrayList( int initialLength )
  {
    _objects = new Object[ initialLength ];
  }

  /**
   * Adds an object to the end of the list.
   * @param o An object to add to the end of the list.
   */
  public final void add( Object o )
  {
    if( _length == _objects.length )
    {
      Object t[] = null;

      if( _objects.length == 0 )
      {
        t = new Object[ 1 ];
      }
      else
      {
        t = new Object[ _objects.length * 2 ];
      }

      for( int i = 0; i < _length; ++i )
      {
        t[ i ] = _objects[ i ];
      }

      _objects = t;
    }
    _objects[ _length++ ] = o;
  }

  /**
   * Inserts an element at a given position in the list.
   */
  public void insert( Object o, int position )
  {
    if( position == _length )
    {
      add( o );
    }
    else if( _length == _objects.length )
    {
      Object t[] = null;

      if( _objects.length == 0 )
      {
        t = new Object[ 1 ];
      }
      else
      {
        t = new Object[ _objects.length * 2 ];
      }

      for( int i = 0; i < position; ++i )
      {
        t[ i ] = _objects[ i ];
      }

      t[ position ] = o;

      for( int i = position; i < _length; ++i )
      {
        t[ i + 1 ] = _objects[ i ];
      }

      _objects = t;
      ++_length;
    }
    else
    {
      for( int i = _length; i > position; --i )
      {
        _objects[ i ] = _objects[ i - 1 ];
      }
      _objects[ position ] = o;
      ++_length;
    }
  }

  /**
   * Inserts an element into a sorted list.
   */
  public void insert( Object o, ObjectComparator c )
  {
    if( _length == 0 )
    {
      add( o );
      return;
    }
    else
    {
      if( c.compare( _objects[ _length - 1 ], o ) >= 0 )
      {
        add( o );
        return;
      }
    }

    if( _length == _objects.length )
    {
      Object t[] = new Object[ _objects.length * 2 ];

      int i = _length - 1;

      for( ; i > 0 ; --i )
      {
        t[ i + 1 ] = _objects[ i ];

        if( c.compare( _objects[ i - 1 ], o ) >= 0 )
        {
          break;
        }
      }

      if( i == 0 ) t[ 1 ] = _objects[ 0 ];

      t[ i ] = o;

      for( int j = 0; j < i; ++j )
      {
        t[ j ] = _objects[ j ];
      }

      _objects = t;

      ++_length;
    }
    else
    {
      int i = _length - 1;

      for( ; i > 0 ; --i )
      {
        _objects[ i + 1 ] = _objects[ i ];

        if( c.compare( _objects[ i - 1 ], o ) >= 0 )
        {
          break;
        }
      }

      if( i == 0 ) _objects[ 1 ] = _objects[ 0 ];

      _objects[ i ] = o;

      ++_length;
    }
  }

  /**
   * Removes an object from the list.
   * @param o An object to remove from the list.
   */
  public void remove( Object o )
  {
    Object e = null;

    for( int i = _length - 1; i >= 0; --i )
    {
      Object t = _objects[ i ];

      _objects[ i ] = e;

      if( t == o )
      {
        --_length;
        break;
      }

      e = t;
    }
  }

  /**
   * Removes an object from the end of the list.
   * @return The last object on the list.
   */
  public final Object remove()
  {
    if( _length > 0 )
    {
      return _objects[ --_length ];
    }
    else
    {
      return null;
    }
  }

  /**
   * Removes all objects from the list.
   */
  public final void removeAll()
  {
    _length = 0;
  }

  /**
   * Gets the object at the specified index from the list.
   * @param i The index of the object to return.
   * @return The object at the specified index in the list.
   */
  public final Object get( int i )
  {
    return _objects[ i ];
  }

  /**
   * Gets the number of number of object in the list.
   * @return The number of number of object in the list.
   */
  public final int length()
  {
    return _length;
  }

  /**
   * Gets the number of number of object in the list.
   * @return The number of number of object in the list.
   */
  public final int size()
  {
    return _length;
  }

  /**
   * Gets the index of an object.
   * returns -1 if not found.
   */
  public final int getIndex( Object o )
  {
    for( int i = 0; i < _length; ++i )
    {
      if( o == _objects[ i ] ) return i;
    }
    return -1;
  }

  /**
   * Set the object at index i.
   * @return The previous object at index i.
   */
  public final Object set( int i, Object o )
  {
    Object t = _objects[ i ];
    _objects[ i ] = o;
    return t;
  }


  class ArrayListIterator implements Iterator
  {
    private int _position;

    public boolean hasNext()
    {
      return _position < _length;
    }

    public Object next()
    {
      return get( _position++ );
    }
  }

  /**
   * Get an iterator on the list
   */
   public Iterator iterator()
   {
     return new ArrayListIterator();
   }

}
