package liaquay.graphics;

public class TextureFactory
{
  public final static float pi=3.1415926535f;
  public final static float deg2rad=pi/180;
  private static float[][] noiseBuffer;
  private static boolean noiseBufferInitialized=false;
  static int minx,maxx,miny,maxy;

  // C O N S T R U C T O R S

    private TextureFactory() {} // Allow no instances

  // E X A M P L E   M A T E R I A L S

    public static ArrayImage SKY(int w, int h, float density)
    {
      int[] colors=new int[2];
      colors[0]=0x003399;
      colors[1]=0xFFFFFF;
      return colorize( PERLIN(w,h, 0.5f, 2.8f*density, 8, 1024),
        RGBColor.makeGradient(colors,1024));
    }

    public static ArrayImage MARBLE(int w, int h, float density)
    {
      int[] colors=new int[3];
      colors[0]=0x111111;
      colors[1]=0x696070;
      colors[2]=0xFFFFFF;
      return colorize( WAVE(w,h,0.5f,0.64f*density,6,1024),
        RGBColor.makeGradient(colors,1024));
    }

    public static ArrayImage WOOD(int w, int h, float density)
    {
      int[] colors=new int[3];
      colors[0]=0x332211;
      colors[1]=0x523121;
      colors[2]=0x996633;

      return colorize( GRAIN(w,h, 0.5f, 3f*density, 3, 8, 1024),
        RGBColor.makeGradient(colors,1024));
    }

    public static ArrayImage RANDOM(int w, int h)
    {
      int nc=(int)random(2,6);
      int[] colors=new int[nc];
      for (int i=0;i<nc;i++)
        colors[i]=RGBColor.random();

      float persistency=random(0.4f,0.9f);
      float density=random(0.5f,3f);
      int samples=(int)random(1,7f);

      return colorize( PERLIN(w,h, persistency, density, samples, 1024),
        RGBColor.makeGradient(colors,1024));
    }

    public static ArrayImage CHECKERBOARD(int w, int h, int cellbits, int oddColor, int evenColor)
    {
      ArrayImage t=new ArrayImage(w,h);

      int pos=0;
      for (int y=0;y<h;y++)
        for (int x=0;x<w;x++)
          t.set( pos++, (((x>>cellbits)+(y>>cellbits))&1)==0 ? evenColor : oddColor );

      return t;
    }

  // B A S E  T Y P E S

    public static ArrayImage PERLIN(int w, int h, float persistency, float density, int samples, int scale)
    {
      initNoiseBuffer();
      ArrayImage t=new ArrayImage(w,h);
      int pos=0;
      float wavelength=(float)((w>h)?w:h)/density;

      for(int y=0;y<h;y++)
        for(int x=0;x<w;x++)
          t.set( pos++, (int)((float)scale*perlin2d(x,y,wavelength,persistency,samples)));
      return t;
    }

    public static ArrayImage WAVE(int w, int h, float persistency, float density, int samples, int scale)
    {
      initNoiseBuffer();
      ArrayImage t=new ArrayImage(w,h);
      int pos=0;
      float wavelength=(float)((w>h)?w:h)/density;

      for(int y=0;y<h;y++)
        for(int x=0;x<w;x++)
          t.set( pos++, (int)((double)scale*(Math.sin(32*perlin2d(x,y,wavelength,persistency,samples))*0.5+0.5)));
      return t;
    }

    public static ArrayImage GRAIN(int w, int h, float persistency, float density, int samples, int levels, int scale)
    // TIP: For wooden textures
    {
      initNoiseBuffer();
      ArrayImage t=new ArrayImage(w,h);
      int pos=0;
      float wavelength=(float)((w>h)?w:h)/density;
      float perlin;

      for(int y=0;y<h;y++)
        for(int x=0;x<w;x++)
        {
          perlin=(float)levels*perlin2d(x,y,wavelength,persistency,samples);
          t.set( pos++, (int)((float)scale*(perlin-(float)(int)perlin)));
        }
      return t;
    }

  // Perlin noise functions

    private static float perlin2d(float x, float y, float wavelength, float persistence, int samples)
    {
      float sum=0;
      float freq=1f/wavelength;
      float amp=persistence;
      float range=0;

      for (int i=0;i<samples;i++)
      {
        sum+=amp*interpolatedNoise(x*freq,y*freq,i);
        range+=amp;
        amp*=persistence;
        freq*=2;
      }
      return crop(sum/persistence*0.5f+0.5f,0,1);
    }

  // Helper methods

    private static float interpolatedNoise(float x, float y, int octave)
    {
      int intx=(int)x;
      int inty=(int)y;
      float fracx=x-(float)intx;
      float fracy=y-(float)inty;

      float i1=interpolate(noise(intx,inty,octave),noise(intx+1,inty,octave),fracx);
      float i2=interpolate(noise(intx,inty+1,octave),noise(intx+1,inty+1,octave),fracx);

      return interpolate(i1,i2,fracy);
    }

    private static float smoothNoise(int x, int y, int o)
    {
      return  (noise(x-1, y-1,o)+noise(x+1, y-1,o)+noise(x-1, y+1,o)+noise(x+1, y+1,o))/16
        +(noise(x-1, y,o)+noise(x+1, y,o)+noise(x, y-1,o)+noise(x, y+1,o))/8
        +noise(x,y,o)/4;
    }

    private static float noise(int x, int y, int octave)
    {
      return noiseBuffer[octave&3][(x+y*57)&8191];
    }

    private static float noise(int seed, int octave)
    {
      int id=octave&3;
          int n =(seed<<13)^seed;

      if (id==0) return (float)(1f- ((n * (n * n * 15731 + 789221) + 1376312589)&0x7FFFFFFF)*0.000000000931322574615478515625f);
      if (id==1) return (float)(1f- ((n * (n * n * 12497 + 604727) + 1345679039)&0x7FFFFFFF)*0.000000000931322574615478515625f);
      if (id==2) return (float)(1f- ((n * (n * n * 19087 + 659047) + 1345679627)&0x7FFFFFFF)*0.000000000931322574615478515625f);
      return (float)(1f- ((n * (n * n * 16267 + 694541) + 1345679501)&0x7FFFFFFF)*0.000000000931322574615478515625f);
    }

    private static void initNoiseBuffer()
    {
      if (noiseBufferInitialized) return;
      noiseBuffer=new float[4][8192];
      for (int octave=0;octave<4;octave++)
        for (int i=0;i<8192;i++)
          noiseBuffer[octave][i]=noise(i,octave);
      noiseBufferInitialized=true;
    }

    private static final float interpolate(float a, float b, float d)
    {
      float f=(1-(float)Math.cos(d*pi))*0.5f;
      return a+f*(b-a);
    }

    public static final int crop(int num, int min, int max)
    {
      return (num<min)?min:(num>max)?max:num;
    }

    public static final float crop(float num, float min, float max)
    {
      return (num<min)?min:(num>max)?max:num;
    }

    public static final float random(float min, float max)
    {
      return (float)(Math.random()*(max-min)+min);
    }

    public static final ArrayImage colorize( ArrayImage image, int[] pal)
    {
      int width = image.getWidth();
      int height = image.getHeight();

      int range= pal.length - 1;

      for (int i=width*height-1;i>=0;i--)
      {
        image.set( i, pal[crop(image.get( i ),0,range)] );
      }

      return image;
    }
}