package liaquay.graphics.concave;

import liaquay.math.vector.*;
import liaquay.graphics.*;

public class Sphere
{
  //
  // Built on construction
  //
  private int _numberOfRings;
  private int _verticesPerRing;
  private Vector3D[] _normals;
  private Vector2D[] _textureCoords;
  private Vector3D _normalNorth = new Vector3D( 0.0, 1.0, 0.0 );
  private Vector3D _normalSouth = new Vector3D( 0.0, -1.0, 0.0 );
  private Vector2D _textureCoordNorth = new Vector2D( 0.5, 0.0 );
  private Vector2D _textureCoordSouth = new Vector2D( 0.5, 1.0 );
  OpaqueSolid3DRenderer _solidRenderer;
  //
  // Set for each call to render
  //
  private Vector3D[] _vertices;
  private Vector3D[] _rotatedNormals;
  private Vector3D _rotatedNormalNorth = new Vector3D( 0.0, 0.0, 0.0 );
  private Vector3D _rotatedNormalSouth = new Vector3D( 0.0, 0.0, 0.0 );
  private Vector3D _vertexNorth = new Vector3D( 0.0, 0.0, 0.0 );
  private Vector3D _vertexSouth = new Vector3D( 0.0, 0.0, 0.0 );
  private ScreenTriangle _triangle;

  public Sphere( OpaqueSolid3DRenderer solidRenderer, int numberOfRings )
  {
    _solidRenderer = solidRenderer;

    _numberOfRings = numberOfRings;
    _verticesPerRing = ( _numberOfRings * 2 ) + 1;

    int numberOfVertices = _numberOfRings * _verticesPerRing;

    _normals = new Vector3D[ numberOfVertices ];
    _textureCoords = new Vector2D[ numberOfVertices ];

    //
    // Set this array up so the space can be used by the renderer.
    //
    _vertices = new Vector3D[ numberOfVertices ];
    _rotatedNormals = new Vector3D[ numberOfVertices ];
    _triangle = new ScreenTriangle();
    _triangle.p1 = new ScreenVertex();
    _triangle.p2 = new ScreenVertex();
    _triangle.p3 = new ScreenVertex();

    int index = 0;

    double ang = Math.PI / (double)_numberOfRings;
    double angY0 = ang;

    for( int y = 0; y < _numberOfRings; ++y )
    {
      double y0 = Math.cos( angY0 );
      double r0 = Math.sin( angY0 );
      double tv = ( 1.0d - y0 ) / 2.0d;

      for( int x = 0; x < _verticesPerRing; ++x )
      {
        double angX0 = ang * (double)x;

        _normals[ index ] = new Vector3D( r0 * Math.sin( angX0 ),
                                          y0,
                                          r0 * Math.cos( angX0 ) );

        double tu = 1.0d - x / ( _numberOfRings * 2.0d );

        _textureCoords[ index ] = new Vector2D( tu, tv );

        //
        // Set up scratch space for the renderer
        //
        _vertices[ index ] = new Vector3D( 0.0, 0.0, 0.0 );
        _rotatedNormals[ index ] = new Vector3D( 0.0, 0.0, 0.0 );

        ++index;
      }
      angY0 += ang;
    }
  }

  public final void render( Object o,
                            ZBufferArrayImage screen,
                            Matrix4x3 combinedTransform,
                            Vector3D center,
                            double ppd,
                            int screenX,
                            int screenY,
                            int screenR,
                            int[] diffuse,
                            int[] specular )
  {
    OpaqueSphere3D opaqueSphere = (OpaqueSphere3D)o;
    Materials materials = opaqueSphere.getOuterMaterials();
    ArrayImageMaterial textureMaterial = (ArrayImageMaterial)materials.getTexturedMaterial();
    ArrayImage texture = textureMaterial.getArrayImage();
    double radius = opaqueSphere.getSphere3D().getRadius();

    //
    // Rotate the normals
    //
    for( int i = 0; i < _normals.length; ++i )
    {
      _normals[ i ].rot( combinedTransform, _rotatedNormals[ i ] );
    }
    _normalSouth.rot( combinedTransform, _rotatedNormalSouth );
    _normalNorth.rot( combinedTransform, _rotatedNormalNorth );

    double centerX = center.get( 0 );
    double centerY = center.get( 1 );
    double centerZ = center.get( 2 );

    //
    // Scale up the sphere
    //
    for( int i = 0; i < _normals.length; ++i )
    {
      Vector3D normal = _rotatedNormals[ i ];

      Vector3D vertex = _vertices[ i ];

      vertex.set( 0, centerX + ( normal.get( 0 ) * radius ) );
      vertex.set( 1, centerY + ( normal.get( 1 ) * radius ) );
      vertex.set( 2, centerZ + ( normal.get( 2 ) * radius ) );
    }

    _vertexNorth.set( 0, centerX + ( _rotatedNormalNorth.get( 0 ) * radius ) );
    _vertexNorth.set( 1, centerY + ( _rotatedNormalNorth.get( 1 ) * radius ) );
    _vertexNorth.set( 2, centerZ + ( _rotatedNormalNorth.get( 2 ) * radius ) );

    _vertexSouth.set( 0, centerX + ( _rotatedNormalSouth.get( 0 ) * radius ) );
    _vertexSouth.set( 1, centerY + ( _rotatedNormalSouth.get( 1 ) * radius ) );
    _vertexSouth.set( 2, centerZ + ( _rotatedNormalSouth.get( 2 ) * radius ) );

    //
    // Render the middle of the sphere
    //
    int width = screen.getWidth();
    int height = screen.getHeight();

    _triangle.material = materials.getPlainMaterial();

    int indexBase = 0;

    double textureWidth = texture.getWidth() - 1.0d;
    double textureHeight = texture.getHeight() - 1.0d;

    for( int y = 0; y < _numberOfRings - 1; ++y )
    {
      for( int x = 0; x < _numberOfRings << 1; ++x )
      {
        int index1 = indexBase + x;
        int index2 = indexBase + _verticesPerRing + x;
        int index3 = indexBase + x+1;
        int index4 = indexBase + _verticesPerRing + x+1;

        Vector3D normal1 = _rotatedNormals[ index1 ];
        Vector3D normal2 = _rotatedNormals[ index2 ];
        Vector3D normal3 = _rotatedNormals[ index3 ];
        Vector3D normal4 = _rotatedNormals[ index4 ];

        if( ( normal1.dot( _vertices[ index1 ] ) < 0 ) ||
            ( normal2.dot( _vertices[ index2 ] ) < 0 ) ||
            ( normal3.dot( _vertices[ index3 ] ) < 0 ) ||
            ( normal4.dot( _vertices[ index4 ] ) < 0 ) )
        {
          {
            int index = index1;

            Vector3D vertex = _vertices[ index ];
            Vector2D textureCoord = _textureCoords[ index ];

            _triangle.n = _rotatedNormals[ index ];

            _triangle.p1.n = _rotatedNormals[ index ];

            _triangle.p1.x = OpaqueSolid3DRenderer.getScreenX( vertex, width, height, ppd );
            _triangle.p1.y = OpaqueSolid3DRenderer.getScreenY( vertex, width, height, ppd );
            _triangle.p1.z = OpaqueSolid3DRenderer.getScreenZ( vertex, width, height, ppd );

            _triangle.p1.tx = (int)(textureCoord.get( 0 ) * textureWidth);
            _triangle.p1.ty = (int)(textureCoord.get( 1 ) * textureHeight);
          }
          {
            int index = index2;

            Vector3D vertex = _vertices[ index ];
            Vector2D textureCoord = _textureCoords[ index ];

            _triangle.p2.n = _rotatedNormals[ index ];

            _triangle.p2.x = OpaqueSolid3DRenderer.getScreenX( vertex, width, height, ppd );
            _triangle.p2.y = OpaqueSolid3DRenderer.getScreenY( vertex, width, height, ppd );
            _triangle.p2.z = OpaqueSolid3DRenderer.getScreenZ( vertex, width, height, ppd );

            _triangle.p2.tx = (int)(textureCoord.get( 0 ) * textureWidth);
            _triangle.p2.ty = (int)(textureCoord.get( 1 ) * textureHeight);
          }
          {
            int index = index3;

            Vector3D vertex = _vertices[ index ];
            Vector2D textureCoord = _textureCoords[ index ];

            _triangle.p3.n = _rotatedNormals[ index ];

            _triangle.p3.x = OpaqueSolid3DRenderer.getScreenX( vertex, width, height, ppd );
            _triangle.p3.y = OpaqueSolid3DRenderer.getScreenY( vertex, width, height, ppd );
            _triangle.p3.z = OpaqueSolid3DRenderer.getScreenZ( vertex, width, height, ppd );

            _triangle.p3.tx = (int)(textureCoord.get( 0 ) * textureWidth);
            _triangle.p3.ty = (int)(textureCoord.get( 1 ) * textureHeight);
          }

          _solidRenderer.renderTrianglePhongTextured( screen,
                                                     _triangle,
                                                     diffuse,
                                                     specular,
                                                     texture );

          {
            int index = index3;

            Vector3D vertex = _vertices[ index ];
            Vector2D textureCoord = _textureCoords[ index ];

            _triangle.n = _rotatedNormals[ index ];

            _triangle.p1.n = _rotatedNormals[ index ];

            _triangle.p1.x = OpaqueSolid3DRenderer.getScreenX( vertex, width, height, ppd );
            _triangle.p1.y = OpaqueSolid3DRenderer.getScreenY( vertex, width, height, ppd );
            _triangle.p1.z = OpaqueSolid3DRenderer.getScreenZ( vertex, width, height, ppd );

            _triangle.p1.tx = (int)(textureCoord.get( 0 ) * textureWidth);
            _triangle.p1.ty = (int)(textureCoord.get( 1 ) * textureHeight);
          }
          {
            int index = index2;

            Vector3D vertex = _vertices[ index ];
            Vector2D textureCoord = _textureCoords[ index ];

            _triangle.p2.n = _rotatedNormals[ index ];

            _triangle.p2.x = OpaqueSolid3DRenderer.getScreenX( vertex, width, height, ppd );
            _triangle.p2.y = OpaqueSolid3DRenderer.getScreenY( vertex, width, height, ppd );
            _triangle.p2.z = OpaqueSolid3DRenderer.getScreenZ( vertex, width, height, ppd );

            _triangle.p2.tx = (int)(textureCoord.get( 0 ) * textureWidth);
            _triangle.p2.ty = (int)(textureCoord.get( 1 ) * textureHeight);
          }
          {
            int index = index4;

            Vector3D vertex = _vertices[ index ];
            Vector2D textureCoord = _textureCoords[ index ];

            _triangle.p3.n = _rotatedNormals[ index ];

            _triangle.p3.x = OpaqueSolid3DRenderer.getScreenX( vertex, width, height, ppd );
            _triangle.p3.y = OpaqueSolid3DRenderer.getScreenY( vertex, width, height, ppd );
            _triangle.p3.z = OpaqueSolid3DRenderer.getScreenZ( vertex, width, height, ppd );

            _triangle.p3.tx = (int)(textureCoord.get( 0 ) * textureWidth);
            _triangle.p3.ty = (int)(textureCoord.get( 1 ) * textureHeight);
          }

          _solidRenderer.renderTrianglePhongTextured( screen,
                                                     _triangle,
                                                     diffuse,
                                                     specular,
                                                     texture );

        }
      }
      indexBase += _verticesPerRing;
    }

    //
    // Render the south section of the sphere
    //
    for( int x = 0; x < _numberOfRings << 1; ++x )
    {
      int index2 = _normals.length - _verticesPerRing + x + 1;
      int index3 = _normals.length - _verticesPerRing + x;

      Vector3D normal2 = _rotatedNormals[ index2 ];
      Vector3D normal3 = _rotatedNormals[ index3 ];


      if( normal2.dot( _vertices[ index2 ] ) < 0 )
      {
        _triangle.n = normal2;

        {
          Vector3D vertex = _vertexSouth;
          Vector2D textureCoord = _textureCoordSouth;

          _triangle.p1.n = _rotatedNormalSouth;

          _triangle.p1.x = OpaqueSolid3DRenderer.getScreenX( vertex, width, height, ppd );
          _triangle.p1.y = OpaqueSolid3DRenderer.getScreenY( vertex, width, height, ppd );
          _triangle.p1.z = OpaqueSolid3DRenderer.getScreenZ( vertex, width, height, ppd );

          _triangle.p1.tx = (int)(textureCoord.get( 0 ) * textureWidth);
          _triangle.p1.ty = (int)(textureCoord.get( 1 ) * textureHeight);
        }
        {
          int index = index2;

          Vector3D vertex = _vertices[ index ];
          Vector2D textureCoord = _textureCoords[ index ];

          _triangle.p2.n = _rotatedNormals[ index ];

          _triangle.p2.x = OpaqueSolid3DRenderer.getScreenX( vertex, width, height, ppd );
          _triangle.p2.y = OpaqueSolid3DRenderer.getScreenY( vertex, width, height, ppd );
          _triangle.p2.z = OpaqueSolid3DRenderer.getScreenZ( vertex, width, height, ppd );

          _triangle.p2.tx = (int)(textureCoord.get( 0 ) * textureWidth);
          _triangle.p2.ty = (int)(textureCoord.get( 1 ) * textureHeight);
        }
        {
          int index = index3;

          Vector3D vertex = _vertices[ index ];
          Vector2D textureCoord = _textureCoords[ index ];

          _triangle.p3.n = _rotatedNormals[ index ];

          _triangle.p3.x = OpaqueSolid3DRenderer.getScreenX( vertex, width, height, ppd );
          _triangle.p3.y = OpaqueSolid3DRenderer.getScreenY( vertex, width, height, ppd );
          _triangle.p3.z = OpaqueSolid3DRenderer.getScreenZ( vertex, width, height, ppd );

          _triangle.p3.tx = (int)(textureCoord.get( 0 ) * textureWidth);
          _triangle.p3.ty = (int)(textureCoord.get( 1 ) * textureHeight);
        }

        _solidRenderer.renderTrianglePhongTextured( screen,
                                                   _triangle,
                                                   diffuse,
                                                   specular,
                                                   texture );
      }
    }

    //
    // Render the north section of the sphere
    //
    for( int x = 0; x < _numberOfRings << 1; ++x )
    {
      int index2 = x + 1;
      int index3 = x;

      Vector3D normal2 = _rotatedNormals[ index2 ];
      Vector3D normal3 = _rotatedNormals[ index3 ];


      if( normal2.dot( _vertices[ index2 ] ) < 0 )
      {
        _triangle.n = normal2;

        {
          Vector3D vertex = _vertexNorth;
          Vector2D textureCoord = _textureCoordNorth;

          _triangle.p1.n = _rotatedNormalNorth;

          _triangle.p1.x = OpaqueSolid3DRenderer.getScreenX( vertex, width, height, ppd );
          _triangle.p1.y = OpaqueSolid3DRenderer.getScreenY( vertex, width, height, ppd );
          _triangle.p1.z = OpaqueSolid3DRenderer.getScreenZ( vertex, width, height, ppd );

          _triangle.p1.tx = (int)(textureCoord.get( 0 ) * textureWidth);
          _triangle.p1.ty = (int)(textureCoord.get( 1 ) * textureHeight);
        }
        {
          int index = index2;

          Vector3D vertex = _vertices[ index ];
          Vector2D textureCoord = _textureCoords[ index ];

          _triangle.p2.n = _rotatedNormals[ index ];

          _triangle.p2.x = OpaqueSolid3DRenderer.getScreenX( vertex, width, height, ppd );
          _triangle.p2.y = OpaqueSolid3DRenderer.getScreenY( vertex, width, height, ppd );
          _triangle.p2.z = OpaqueSolid3DRenderer.getScreenZ( vertex, width, height, ppd );

          _triangle.p2.tx = (int)(textureCoord.get( 0 ) * textureWidth);
          _triangle.p2.ty = (int)(textureCoord.get( 1 ) * textureHeight);
        }
        {
          int index = index3;

          Vector3D vertex = _vertices[ index ];
          Vector2D textureCoord = _textureCoords[ index ];

          _triangle.p3.n = _rotatedNormals[ index ];

          _triangle.p3.x = OpaqueSolid3DRenderer.getScreenX( vertex, width, height, ppd );
          _triangle.p3.y = OpaqueSolid3DRenderer.getScreenY( vertex, width, height, ppd );
          _triangle.p3.z = OpaqueSolid3DRenderer.getScreenZ( vertex, width, height, ppd );

          _triangle.p3.tx = (int)(textureCoord.get( 0 ) * textureWidth);
          _triangle.p3.ty = (int)(textureCoord.get( 1 ) * textureHeight);
        }

        _solidRenderer.renderTrianglePhongTextured( screen,
                                                   _triangle,
                                                   diffuse,
                                                   specular,
                                                   texture );
      }
    }



  }
}