package liaquay.graphics.convex;

import java.awt.*;
import liaquay.math.vector.*;
import liaquay.graphics.*;
import liaquay.collection.*;

/**
 * This implementation is not thread safe.
 */
public class ConvexGraphics3D implements Graphics3D
{
  //
  // Renderes for the different types of object.
  //
  private OpaqueSolid3DRenderer _opaqueSolid3DRenderer = new OpaqueSolid3DRenderer();
  private OpaqueSphere3DRenderer _opaqueSphere3DRenderer = new OpaqueSphere3DRenderer();

  //
  // private variables
  //
  private Image _image;
  private int _width;
  private int _height;
  private Transform _inverseCamera = new Transform();
  private double _ppd;
  private ViewListEntry[] _viewListEntries = new ViewListEntry[ 300 ];
  private int _length = 0;
  private Selection _selection;

  /**
   *
   */
  public ConvexGraphics3D( Image image, Selection selection )
  {
    _image = image;
    _width = image.getWidth(null);
    _height = image.getHeight(null);
    _selection = selection;
  }

  /**
   * Prepares the renderer for new a new scene.
   */
  public void beginScene( Camera3D camera )
  {
    camera.getTransform().inverse( _inverseCamera );
    _ppd = camera.getDistanceToFocalPoint();
    clear();
  }

  protected void added( ViewListEntry viewListEntry,
                        SelectionListener selectionListener )
  {
  }

  /**
   * Adds an opaque solid to the scene.
   */
  public void add( OpaqueSolid3D solid,
                   Transform transform,
                   SelectionListener selectionListener )
  {
    ViewListEntry viewListEntry = getViewListEntry();

    viewListEntry.initialise( _inverseCamera,
                              transform,
                              solid.getSolid3D().getCenter(),
                              solid.getSolid3D().getRadius(),
                              _ppd,
                              solid,
                              _opaqueSolid3DRenderer,
                              _width,
                              _height,
                              selectionListener );

    added( viewListEntry, selectionListener );

    if( !viewListEntry.onScreen( _width, _height ) )
    {
      discard( viewListEntry );
    }
  }

  public void add( TranslucentSolid3D solid,
                   Transform transform,
                   SelectionListener selectionListener )
  {
  }

  /**
   * Adds an opaque shpere to the scene.
   */
  public void add( OpaqueSphere3D sphere,
                   Transform transform,
                   SelectionListener selectionListener )
  {
    ViewListEntry viewListEntry = getViewListEntry();

    viewListEntry.initialise( _inverseCamera,
                              transform,
                              sphere.getSphere3D().getCenter(),
                              sphere.getSphere3D().getRadius(),
                              _ppd,
                              sphere,
                              _opaqueSphere3DRenderer,
                              _width,
                              _height,
                              selectionListener );

    added( viewListEntry, selectionListener );

    if( !viewListEntry.onScreen( _width, _height ) )
    {
      discard( viewListEntry );
    }
  }

  public void add( TranslucentSphere3D sphere,
                   Transform transform,
                   SelectionListener selectionListener )
  {
  }

  public void add( OpaquePlane3D plane,
                   Transform transform,
                   SelectionListener selectionListener )
  {
  }

  /**
   * Compeletes the scene.
   */
  public void endScene()
  {
    sort();

    Graphics graphics = _image.getGraphics();

    graphics.setColor( Color.black );
    graphics.fillRect( 0, 0, _width, _height );

    graphics.translate( _width/2, _height/2 );

    renderScene( graphics );

    graphics.translate( -_width/2, -_height/2 );
  }

  public void renderScene( Graphics graphics )
  {
    for( int i = 0; i < _length; ++i )
    {
      ViewListEntry e = _viewListEntries[ i ];
      e.render( graphics, _ppd, _width, _height );
    }
  }

  /**
   * Get an image representing the rendered scene.
   */
  public Image getImage()
  {
    return _image;
  }

  public Selection getSelection()
  {
    return _selection;
  }

  /**
   * Deliver a pick event to the graphics engine.
   * This will result in the appropriate selection listener being activated.
   */
  public void pick( int x, int y )
  {
    if( _selection == null ) return;

    int pickX = x -_width/2;
    int pickY = y - _height/2;

    for( int i = _length-1; i >= 0; --i )
    {
      ViewListEntry viewListEntry = _viewListEntries[ i ];

      if( viewListEntry.pick( pickX, pickY, _ppd ) )
      {
        SelectionListener selectionListener = viewListEntry.getSelectionListener();

        _selection.select( selectionListener );

        return;
      }
    }

    _selection.select( null );
  }

  private final void add( ViewListEntry o )
  {
    if( _length == _viewListEntries.length )
    {
      ViewListEntry t[] = new ViewListEntry[ _viewListEntries.length * 2 ];
      for( int i = 0; i < _viewListEntries.length; ++i )
      {
        t[ i ] = _viewListEntries[ i ];
      }
      _viewListEntries = t;
    }
    _viewListEntries[ _length++ ] = o;
  }

  private final ViewListEntry getViewListEntry()
  {
    ViewListEntry o = _viewListEntries[ _length ];

    if( o == null )
    {
      o = new ViewListEntry();

      add( o );

      return o;
    }
    else
    {
      ++_length;
      return o;
    }
  }

  private final void discard( ViewListEntry e )
  {
    // Remove this after testing...
    if( e != _viewListEntries[ _length - 1 ] )
    {
      throw new RuntimeException();
    }

    --_length;
  }

  public final void clear()
  {
    _length = 0;
  }

  public final int getLength()
  {
    return _length;
  }

  public final double getPpd()
  {
    return _ppd;
  }

  public final int getWidth()
  {
    return _width;
  }

  public final int getHeight()
  {
    return _height;
  }

  public final ViewListEntry getViewListEntry( int index )
  {
    return _viewListEntries[ index ];
  }

  public final void sort()
  {
    quickSort( 0, _length -1 );
  }

  private final void quickSort( int f, int l )
  {
    if( f == l )
    {
    }
    else if( l - f < 9 )
    {
      SelectionSort( f, l );
    }
    else
    {
      int pivotIndex = partition( f, l );
      quickSort( f, pivotIndex - 1 );
      quickSort( pivotIndex + 1, l );
    }
  }

  private final int partition( int f, int l )
  {
    int fi = f;
    int li = l;

    ViewListEntry pivot = _viewListEntries[ f ];

    ++f;
    while( f < l )
    {
      while( ( f < l ) && ( _viewListEntries[ f ].compareTo( pivot ) <= 0  ) )
      {
        ++f;
      }

      while( ( f <= l ) && ( _viewListEntries[ l ].compareTo( pivot ) >= 0  ) )
      {
        --l;
      }

      if( f < l )
      {
        ViewListEntry temp = _viewListEntries[ f ];
        _viewListEntries[ f ] = _viewListEntries[ l ];
        _viewListEntries[ l ] = temp;
      }
    }

    ViewListEntry temp = _viewListEntries[ fi ];
    _viewListEntries[ fi ] = _viewListEntries[ l ];
    _viewListEntries[ l ] = temp;

    return l;
  }

  private final void SelectionSort( int f, int l )
  {
    for( int i = f; i <= l; ++i )
    {
      ViewListEntry key = _viewListEntries[ i ];

      for( int j = i + 1; j <= l; ++j )
      {
        ViewListEntry test = _viewListEntries[ j ];

        if( key.compareTo( test ) > 0 )
        {
          _viewListEntries[ i ] = test;
          _viewListEntries[ j ] = key;
          key = test;
        }
      }
    }
  }
}