package liaquay.graphics.raytrace;

import java.awt.*;

import liaquay.collection.*;
import liaquay.math.vector.*;
import liaquay.graphics.*;

public class RayBox
{
  private ArrayList _traceables = new ArrayList();
  private Colour _ambientColour = new Colour(1.0f,1.0f,1.0f);
  private double _refractiveIndex = 1;
  private int _maxBounces = 20;

  public RayBox( int maxBounces )
  {
    _maxBounces = maxBounces;
  }

  public void add( Traceable t )
  {
    _traceables.add( t );
  }

  public void removeAll()
  {
    _traceables.removeAll();
  }

  public Colour getAmbientColour()
  {
    return _ambientColour;
  }

  public Colour getReflectiveColour( Vector3D intersection,
                                     Vector3D direction,
                                     Vector3D normal,
                                     int depth )
  {
    // Component in direction of normal
    double c = -normal.dot( direction );

    Vector3D d = direction.add( normal.mul( c * 2 ) );

    return trace( intersection, d, depth );
  }

  private Vector3D getRefractedDirection( Vector3D direction,
                                          Vector3D normal,
                                          double refractiveIndex )
  {
    double v = normal.dot( direction );
    boolean entering = v <= 0;

    double k = entering ? _refractiveIndex /  refractiveIndex :
                           refractiveIndex / _refractiveIndex;
    double kk = k * k;
    double vv = v * v;
    Vector3D vp = direction.sub( normal.mul( v ) ).mul( k );
    double f = 1 - ( kk * ( 1 - vv ) );

    if( f < 0 )
    {
      return null;
    }

    Vector3D vn = normal.mul( Math.sqrt( f ) );
    return entering ? vp.sub( vn ) : vp.add( vn );
  }

  public Colour getRefractedColour( Vector3D intersection,
                                    Vector3D direction,
                                    Vector3D normal,
                                    double refractiveIndex,
                                    int depth )
  {
    Vector3D d = getRefractedDirection( direction,
                                        normal,
                                        refractiveIndex );

    if( d == null )
    {
      return getReflectiveColour( intersection,
                                  direction,
                                  normal,
                                  depth );
    }
    return trace( intersection, d, depth );
  }

  public Colour trace( Vector3D start,
                       Vector3D direction,
                       int depth )
  {
    double minimumDistanceSquared = 0.0;
    Traceable traceable = null;
    Vector3D intersection = null;

    if( depth < _maxBounces )
    {
      for( int i = 0; i < _traceables.size(); ++i )
      {
        Traceable t = (Traceable)_traceables.get( i );
        Vector3D x = t.intersection( start, direction );

        if( x != null )
        {
          double d = x.sub( start ).lengthSquared();

          if( ( traceable == null ) || ( d < minimumDistanceSquared ) )
          {
            if( d > 0.0001 )
            {
              minimumDistanceSquared = d;
              traceable = t;
              intersection = x;
            }
          }
        }
      }
    }

    if( traceable != null )
    {
      return traceable.trace( intersection,
                              direction,
                              this,
                              depth + 1 );
    }
    else
    {
      return getAmbientColour();
    }
  }

  public void render( Graphics graphics,
                      int originX,
                      int originY,
                      int imageWidth,
                      int imageHeight,
                      int oversample,
                      double ppd )
  {
    int width = imageWidth * oversample;
    int height = imageHeight * oversample;

    int widthBy2 = width / 2;
    int heightBy2 = height / 2;

    float scale = 1.0f / (float)( oversample * oversample );

    Vector3D start = new Vector3D( 0, 0, 0 );

    for( int x = 0; x < imageWidth; ++x )
    {
      for( int y = 0; y < imageHeight; ++y )
      {
        float red =0, green = 0, blue = 0;

        for( int xx = x * oversample; xx < (x+1) * oversample; ++xx )
        {
          for( int yy = y * oversample; yy < (y+1) * oversample; ++yy )
          {
            Vector3D direction = new Vector3D( (double)( xx - widthBy2 ) / (double)width,
                                               -(double)( yy - heightBy2 ) / (double)width,
                                               ppd ).normalize();

            Colour colour = trace( start, direction, 0 );

            red += colour.getRed();
            blue += colour.getBlue();
            green += colour.getGreen();
          }
        }

        red *= scale;
        green *= scale;
        blue *= scale;

        // Prevent burn out...
        float max = red;
        if( green > max ) max = green;
        if( blue > max ) max = blue;
        if( max > 1.0 )
        {
          red /= max;
          blue /= max;
          green /= max;
        }

        Color color = new Color( (float)red,
                                 (float)green,
                                 (float)blue );

        graphics.setColor( color );
        graphics.fillRect(  x + originX, y + originY, 1, 1 );
      }
    }
  }

  public static Colour compose( Colour reflective,
                                Colour emissive,
                                Colour incident,
                                Colour ambient,
                                float reflectivity,
                                double cosAngleToNormal )
  {
    float red = reflective.getRed() *
             ( ( incident.getRed() * reflectivity ) +
               ( ambient.getRed() * ( 1.0f - reflectivity ) ) );

    float green = reflective.getGreen() *
               ( ( incident.getGreen() * reflectivity ) +
                 ( ambient.getGreen() * ( 1.0f - reflectivity ) ) );

    float blue = reflective.getBlue() *
              ( ( incident.getBlue() * reflectivity ) +
                ( ambient.getBlue() * ( 1.0f - reflectivity ) ) );

    float f = 0.5f + ( (float)cosAngleToNormal * 0.5f );

    red = (red * f) + emissive.getRed();
    green = (green * f) + emissive.getGreen() ;
    blue = (blue * f) + emissive.getBlue();

    return new Colour( red, green, blue );
  }

  public static Colour compose( Colour reflective,
                                Colour emissive,
                                Colour incidentReflection,
                                Colour incidentRefraction,
                                Colour ambient,
                                float reflectivity,
                                double cosAngleToNormal,
                                Colour refractive,
                                float translucency )
  {
    float f = 0.5f + ( (float)cosAngleToNormal * 0.5f );

    float redReflective =
             f * ( reflective.getRed() *
                   ( ( incidentReflection.getRed() * reflectivity ) +
                   ( ambient.getRed() * ( 1.0f - reflectivity ) ) ) );

    float greenReflective =
             f * ( reflective.getGreen() *
                   ( ( incidentReflection.getGreen() * reflectivity ) +
                   ( ambient.getGreen() * ( 1.0f - reflectivity ) ) ) );

    float blueReflective =
             f * ( reflective.getBlue() *
                   ( ( incidentReflection.getBlue() * reflectivity ) +
                   ( ambient.getBlue() * ( 1.0f - reflectivity ) ) ) );


    float red = ( translucency * refractive.getRed() * incidentRefraction.getRed() ) +
           ( ( 1 - translucency ) * redReflective ) +
           emissive.getRed();

    float green = ( translucency * refractive.getGreen() * incidentRefraction.getGreen()  ) +
             ( ( 1 - translucency ) * greenReflective ) +
             emissive.getGreen() ;

    float blue = ( translucency * refractive.getBlue() * incidentRefraction.getBlue()  ) +
            ( ( 1 - translucency ) * blueReflective ) +
            emissive.getBlue();

    return new Colour( red, green, blue );
  }
}