package liaquay.graphics.raytrace;

import java.awt.*;
import liaquay.math.vector.*;
import liaquay.graphics.*;

public class RayTraceGraphics3D implements Graphics3D
{
  private Image _image;
  private RayBox _raybox;
  private int _maxBounces;
  private int _oversample;
  private Transform _inverseCamera = null;
  private double _ppd;

  /**
   *
   */
  public RayTraceGraphics3D( Image image,
                             int maxBounces,
                             int oversample )
  {
    _image = image;
    _maxBounces = maxBounces;
    _oversample = oversample;
    _raybox = new RayBox( maxBounces );
  }

  /**
   * Prepares the renderer for new a new scene.
   */
  public void beginScene( Camera3D camera )
  {
    _raybox.removeAll();
    _inverseCamera = camera.getTransform().inverse();
    _ppd = camera.getDistanceToFocalPoint();
  }

  /**
   * Adds an opaque solid to the scene.
   */
  public void add( OpaqueSolid3D solid,
                   Transform transform,
                   SelectionListener selectionListener )
  {
    Transform combinedTransform = transform.mul( _inverseCamera );

    Solid3D transformedSolid = solid.getSolid3D().mul( new Matrix4x3( combinedTransform ) );

    TraceableSolid t = new TraceableSolid( transformedSolid, solid.getOuterMaterials() );

    _raybox.add( t );
  }

  public void add( TranslucentSolid3D solid,
                   Transform transform,
                   SelectionListener selectionListener )
  {
    Transform combinedTransform = transform.mul( _inverseCamera );

    Solid3D transformedSolid = solid.getSolid3D().mul( new Matrix4x3( combinedTransform ) );

    TraceableTranslucentSolid t = new TraceableTranslucentSolid(
      transformedSolid,
      solid.getOuterMaterials(),
      solid.getRefractiveIndex(),
      solid.getTranslucency() );

    _raybox.add( t );
  }

  /**
   * Adds an opaque shpere to the scene.
   */
  public void add( OpaqueSphere3D sphere,
                   Transform transform,
                   SelectionListener selectionListener )
  {
    Transform combinedTransform = transform.mul( _inverseCamera );

    Sphere3D transformedSphere = sphere.getSphere3D().mul( new Matrix4x3( combinedTransform ) );

    TraceableSphere t = new TraceableSphere( transformedSphere,
                                             sphere.getOuterMaterials() );

    _raybox.add( t );
  }

  public void add( TranslucentSphere3D sphere,
                   Transform transform,
                   SelectionListener selectionListener )
  {
    Transform combinedTransform = transform.mul( _inverseCamera );

    Sphere3D transformedSphere = sphere.getSphere3D().mul( new Matrix4x3( combinedTransform ) );

    TraceableTranslucentSphere t =
      new TraceableTranslucentSphere( transformedSphere,
                                      sphere.getOuterMaterials(),
                                      sphere.getRefractiveIndex(),
                                      sphere.getTranslucency() );

    _raybox.add( t );
  }

  public void add( OpaquePlane3D plane,
                   Transform transform,
                   SelectionListener selectionListener )
  {
    Transform combinedTransform = transform.mul( _inverseCamera );

    MappedPlane3D transformedPlane = plane.getMappedPlane3D().mul( new Matrix4x3( combinedTransform ) );

    TraceablePlane t = new TraceablePlane( transformedPlane,
                                           plane.getUpperMaterials() );

    _raybox.add( t );
  }

  /**
   * Compeletes the scene.
   */
  public void endScene()
  {
    _raybox.render( _image.getGraphics(),
                    0,
                    0,
                    _image.getWidth(null),
                    _image.getHeight(null),
                    _oversample,
                    _ppd );
  }

  /**
   * Get an image representing the rendered scene.
   */
  public Image getImage()
  {
    return _image;
  }

  /**
   * Deliver a pick event to the graphics engine.
   * This will result in the appropriate selection listener being activated.
   * Only use this method after a call to endScene().
   */
  public void pick( int x, int y )
  {
  }

}