package liaquay.graphics.raytrace;

import liaquay.collection.*;
import liaquay.math.vector.*;
import liaquay.graphics.*;

public class TraceableSolid implements Traceable
{
  private ArrayList _faces = new ArrayList();
  private Solid3D _solid;

  public TraceableSolid( Solid3D solid, Materials[] materials )
  {
    _solid = solid;

    Solid3DFace[] faces = solid.getFaces();

    for( int i = faces.length - 1; i >= 0 ; --i )
    {
      Solid3DFace face = faces[ i ];

      Materials m = materials[ i ];

      add( face, m );
    }
  }

  private void add( Solid3DFace face, Materials materials )
  {
    int[] v = face.getVertices();

    Vector3D faceNormal = face.calculateNormal( _solid.getVertices() );
    Vector3D origin = _solid.getVertices()[ v[ 0 ] ];
    Vector3D axisX = _solid.getVertices()[ v[ 1 ] ].sub( origin ).normalize();
    Vector3D axisY = faceNormal.cross( axisX );

    MappedPlane3D plane = new MappedPlane3D( origin,
                                             origin.add( axisX ),
                                             origin.add( axisY ),
                                             false );


    double[][] points = new double[ v.length ][ 2 ];

    for( int i = 0; i < v.length; ++i )
    {
      Vector3D t1 = _solid.getVertices()[ v[ i ] ];

      double x = plane.getMappedX( t1 );
      double y = plane.getMappedY( t1 );

      points[i][0] = x;
      points[i][1] = y;
    }

    TraceablePolygon poly =
      new TraceablePolygon( plane,
                            points,
                            materials );

    _faces.add( poly );
  }


  public Colour trace( Vector3D intersection,
                       Vector3D direction,
                       RayBox box,
                       int depth )
  {
    for( int i = 0; i < _faces.size(); ++i )
    {
      TraceablePolygon poly = (TraceablePolygon)_faces.get( i );

      if( poly.contains( intersection ) )
      {
        return poly.trace( intersection,
                           direction,
                           box,
                           depth );
      }
    }

    // Hmm?
    return null;
  }


  public Vector3D intersection( Vector3D start,
                                Vector3D direction )
  {
    // Vector from the start of the ray to the center of the
    // volume sphere.
    Vector3D t1 = _solid.getCenter().sub( start );

    // Distance from the start of the ray to a plane
    // perpendicular to the ray, through center of volume sphere.
    double t2 = t1.dot( direction );

    // Point of collision of line and a plane perpendicular
    // to the ray, through center of sphere.
    Vector3D p = start.add( direction.mul( t2 ) );

    // Distance from the center of the sphere to the
    // collision squared.
    double dd = p.sub( _solid.getCenter() ).lengthSquared();

    // Test there is a collision with the volume sphere.
    if( dd < _solid.getRadiusSquared() )
    {
      Vector3D q = null;
      double minimumDistanceSquared = 0.0;

      for( int i = 0; i < _faces.size(); ++i )
      {
        TraceablePolygon poly = (TraceablePolygon)_faces.get( i );

        Vector3D x = poly.intersection( start, direction );

        if( x != null )
        {
          double d = x.sub( start ).lengthSquared();

          if( ( q == null ) || ( d < minimumDistanceSquared ) )
          {
            if( d > 0.0001 )
            {
              q = x;
              minimumDistanceSquared = d;
            }
          }
        }
      }

      return q;
    }

    return null;
  }
}