package liaquay.graphics.raytrace;

import liaquay.collection.*;
import liaquay.math.vector.*;
import liaquay.graphics.*;

public class TraceableTranslucentSolid implements Traceable
{
  private static final double TOL = 0.001;
  private ArrayList _faces = new ArrayList();
  private Solid3D _solid;
  private double _refractiveIndex;
  private float _translucency;

  public TraceableTranslucentSolid( Solid3D solid,
                                    Materials[] materials,
                                    double refractiveIndex,
                                    float translucency  )
  {
    _solid = solid;
    _refractiveIndex = refractiveIndex;
    _translucency = translucency;

    Solid3DFace[] faces = solid.getFaces();

    for( int i = faces.length - 1; i >= 0 ; --i )
    {
      Solid3DFace face = faces[ i ];

      Materials m = materials[ i ];

      add( face, m );
    }
  }

  private void add( Solid3DFace face, Materials materials )
  {
    int[] v = face.getVertices();

    Vector3D faceNormal = face.calculateNormal( _solid.getVertices() );
    Vector3D origin = _solid.getVertices()[ v[ 0 ] ];
    Vector3D axisX = _solid.getVertices()[ v[ 1 ] ].sub( origin ).normalize();
    Vector3D axisY = faceNormal.cross( axisX );

    MappedPlane3D plane = new MappedPlane3D( origin,
                                             origin.add( axisX ),
                                             origin.add( axisY ),
                                             false );


    double[][] points = new double[ v.length ][ 2 ];

    for( int i = 0; i < v.length; ++i )
    {
      Vector3D t1 = _solid.getVertices()[ v[ i ] ];

      double x = plane.getMappedX( t1 );
      double y = plane.getMappedY( t1 );

      points[i][0] = x;
      points[i][1] = y;
    }

    TraceableTranslucentPolygon poly =
      new TraceableTranslucentPolygon( plane,
                            points,
                            materials );

    _faces.add( poly );
  }


  public Colour trace( Vector3D intersection,
                       Vector3D direction,
                       RayBox box,
                       int depth )
  {
    for( int i = 0; i < _faces.size(); ++i )
    {
      TraceableTranslucentPolygon poly = (TraceableTranslucentPolygon)_faces.get( i );

      if( poly.contains( intersection, TOL ) )
      {
        return poly.trace( intersection,
                           direction,
                           box,
                           depth,
                           _refractiveIndex,
                           _translucency );
      }
    }

    return new Colour( 0,0,0 );
  }

  public Vector3D intersection( Vector3D start,
                                Vector3D direction )
  {
    // Vector from the start of the ray to the center of the
    // volume sphere.
    Vector3D t1 = _solid.getCenter().sub( start );

    // Distance from the start of the ray to a plane
    // perpendicular to the ray, through center of volume sphere.
    double t2 = t1.dot( direction );

    // Point of collision of line and a plane perpendicular
    // to the ray, through center of sphere.
    Vector3D p = start.add( direction.mul( t2 ) );

    // Distance from the center of the sphere to the
    // collision squared.
    double dd = p.sub( _solid.getCenter() ).lengthSquared();

    // Test there is a collision with the volume sphere.
    if( dd < _solid.getRadiusSquared() )
    {
      ArrayList intersections = getIntersections( start, direction );

      if( intersections.size() > 0 )
      {
        TranslucentPolygonIntersection i0 = (TranslucentPolygonIntersection)intersections.get( 0 );

        if( !i0.getPolygon().contains( start, TOL ) )
        {
          return i0.getIntersection();
        }
        else if( i0.getPolygon().entering( direction ) )
        {
          //
          // The line is entering the solid
          //

          // A count of entry/exit to polygon
          int n = 1;
          int i;

          for( i = 1; i < intersections.size(); ++i )
          {
            TranslucentPolygonIntersection intersection = (TranslucentPolygonIntersection)intersections.get( i );
            n += intersection.getPolygon().entering( direction ) ? +1 : -1;
            if( n == 0 ) break;
          }

          if( i < intersections.size() )
          {
            TranslucentPolygonIntersection intersection = (TranslucentPolygonIntersection)intersections.get( i );

            return intersection.getIntersection();
          }
          else
          {
            return null;
          }
        }
        else
        {
          //
          // The line is exiting the solid
          //

          for( int i = 0; i < intersections.size(); ++i )
          {
            TranslucentPolygonIntersection intersection = (TranslucentPolygonIntersection)intersections.get( i );
            Vector3D x = intersection.getIntersection();
            if( x.sub( start ).lengthSquared() > (TOL*TOL) )
            {
              return x;
            }
          }
        }

      }
      else
      {
        return null;
      }
    }

    return null;
  }

  /**
   * Get the intersections including those on the start point in order of distance
   */
  private ArrayList getIntersections( Vector3D start,
                                      Vector3D direction )
  {
    ArrayList list = new ArrayList();

    for( int i = 0; i < _faces.size(); ++i )
    {
      TraceableTranslucentPolygon poly = (TraceableTranslucentPolygon)_faces.get( i );

      // Find intersections with the faces including those on the start point
      Vector3D intersection = poly.halfLineIntersection( start, direction, -TOL );

      if( intersection != null )
      {
        list.add( new TranslucentPolygonIntersection( poly, intersection ) );
      }
    }

    // Order the intersections in terms of distance from start
    SortableObjectArray sortable = new SortableObjectArray( list, new IntersectionComparator( start ) );

    QuickSorter.SORTER.sort( sortable );


    return list;
  }
}

