package liaquay.math.vector;

public class MappedPlane3D
{
  private Vector3D _origin;
  private Vector3D _pointX;
  private Vector3D _pointY;
  private boolean _invertNormal;

  private Vector3D _axisX;
  private Vector3D _axisY;


  // The normal to the plane as a unit vector.
  private Vector3D _normal;

  // The distance from the origin to the plane in the direction of the normal.
  private double _distance;

  public MappedPlane3D( Vector3D origin,
                        Vector3D pointX,
                        Vector3D pointY,
                        boolean invertNormal )
  {
    _origin = origin;
    _pointX = pointX;
    _pointY = pointY;
    _invertNormal = invertNormal;

    _axisX = pointX.sub( origin );
    _axisY = pointY.sub( origin );
    Vector3D t3 = _axisX.cross( _axisY );
    Vector3D t4 = t3.normalize();
    if( invertNormal )
    {
      _normal = t4.invert();
    }
    else
    {
      _normal = t4;
    }
    _distance = _normal.dot( origin );
  }

  /**
   * Gets the normal to the plane.
   * @return The normal to the plane.
   */
  public Vector3D getNormal()
  {
    return _normal;
  }

  public Vector3D getAxisX()
  {
    return _axisX;
  }

  public Vector3D getAxisY()
  {
    return _axisY;
  }

  public Vector3D getOrigin()
  {
    return _origin;
  }

  /**
   * Gets the distance from the origin to the plane in the direction of the normal.
   * @return The distance from the origin to the plane in the direction of the normal.
   */
  public double getDistance()
  {
    return _distance;
  }

  /**
   * Gets the point of intersection of the plane and a line.
   * @param line a line in 3-dimentional space.
   * @return The point of intersection of the plane and the line.
   */
  public Vector3D intersection( Line3D line )
  {
    return intersection( line.getPoint(), line.getDirection() );
  }

  /**
   * Gets the point of intersection of the plane and a line.
   * @param start A point on the line.
   * @param direction The direction of the line.
   * @return The point of intersection of the plane and the line.
   */
  public Vector3D intersection( Vector3D start,
                                Vector3D direction )
  {
    double t1 = _normal.dot( direction );

    // is the direction parallel to the plane?
    if( ( t1 < 0.0001 ) && ( t1 > -0.0001 ) )
    {
      return null;
    }

    double t2 = start.dot( _normal );

    double t3 = ( _distance - t2 ) / t1;

    // Point of collision of line and the plane.
    return start.add( direction.mul( t3 ) );
  }

  /**
   * Gets the point of intersection of the plane and a half-line.
   * @param start A point on the line.
   * @param direction The direction of the line.
   * @return The point of intersection of the plane and the line.
   */
  public Vector3D halfLineIntersection( Vector3D start,
                                        Vector3D direction )
  {
    double t1 = _normal.dot( direction );

    // is the direction parallel to the plane?
    if( ( t1 < 0.0001 ) && ( t1 > -0.0001 ) )
    {
      return null;
    }

    double t2 = start.dot( _normal );

    double t3 = ( _distance - t2 ) / t1;

    // Point of collision of line and the plane.
    Vector3D p = start.add( direction.mul( t3 ) );

    Vector3D t4 = p.sub( start );

    double t5 = t4.dot( direction );

    if( t5 > 0 )
    {
      return p;
    }
    else
    {
      return null;
    }
  }

  /**
   * Gets the point of intersection of the plane and a half-line.
   * @param start A point on the line.
   * @param direction The direction of the line. a Unit vector!
   * @return The point of intersection of the plane and the line.
   * @param minDistance Mininmum distance of intersection from start. Use -ve to include start.
   */
  public Vector3D halfLineIntersection( Vector3D start,
                                        Vector3D direction,
                                        double minDistance )
  {
    double t1 = _normal.dot( direction );

    // is the direction parallel to the plane?
    if( ( t1 < 0.0001 ) && ( t1 > -0.0001 ) )
    {
      return null;
    }

    double t2 = start.dot( _normal );

    double t3 = ( _distance - t2 ) / t1;

    // Point of collision of line and the plane.
    Vector3D p = start.add( direction.mul( t3 ) );

    Vector3D t4 = p.sub( start );

    double t5 = t4.dot( direction );

    if( t5 > 0 )
    {
      if( minDistance > 0 )
      {
        if( t5 > minDistance )
        {
          return p;
        }
        else
        {
          return null;
        }
      }
      else
      {
        return p;
      }
    }
    else
    {
      if( minDistance < 0 )
      {
        if( t5 > minDistance )
        {
          return p;
        }
        else
        {
          return null;
        }
      }
      else
      {
        return null;
      }
    }
  }


  /**
   * Tests if the specified point lies on this plane.
   * @param point A point in 3-dimentional space.
   * @param maxDistance The maximum allowable distance of the point from the plane.
   * @return true if the point lies on this plane.
   */
  public boolean contains( Vector3D point, double maxDistance )
  {
    double t1 = point.dot( _normal ) - _distance;
    return ( t1 < maxDistance ) && ( t1 > -maxDistance );
  }

  public double getMappedX( Vector3D point )
  {
    return point.sub( _origin ).dot( _axisX );
  }

  public double getMappedY( Vector3D point )
  {
    return point.sub( _origin).dot( _axisY );
  }

  public Vector2D getMapped( Vector3D point )
  {
    return new Vector2D( getMappedX( point ), getMappedY( point ) );
  }

  public MappedPlane3D mul( Matrix4x3 m )
  {
    Vector3D origin = _origin.mul( m );
    Vector3D pointX = _pointX.mul( m );
    Vector3D pointY = _pointY.mul( m );
    boolean invertNormal = _invertNormal;

    return new MappedPlane3D( origin, pointX, pointY, invertNormal );
  }

  public boolean above( Vector3D point )
  {
    return point.dot( _normal ) > _distance;
  }
}