package liaquay.math.vector;

/**
 * Representation of a plane in 3-dimentional space.
 */
public class Plane3D
{
  // The normal to the plane as a unit vector.
  private Vector3D _normal;

  // The distance from the origin to the plane in the direction of the normal.
  private double _distance;

  /**
   * Constructs a plane in 3-dimentional space.
   * @param normal The normal to the plane. Must be a unit vector.
   * @param distance he distance from the origin to the plane in the direction of the normal.
   */
  public Plane3D( Vector3D normal, double distance )
  {
    _normal = normal;
    _distance = distance;
  }

  /**
   * Gets the normal to the plane.
   * @return The normal to the plane.
   */
  public Vector3D getNormal()
  {
    return _normal;
  }

  /**
   * Gets the distance from the origin to the plane in the direction of the normal.
   * @return The distance from the origin to the plane in the direction of the normal.
   */
  public double getDistance()
  {
    return _distance;
  }

  /**
   * Gets the point of intersection of the plane and a line.
   * @param line a line in 3-dimentional space.
   * @return The point of intersection of the plane and the line.
   */
  public Vector3D intersection( Line3D line )
  {
    return intersection( line.getPoint(), line.getDirection() );
  }

  /**
   * Gets the point of intersection of the plane and a line.
   * @param start A point on the line.
   * @param direction The direction of the line.
   * @return The point of intersection of the plane and the line.
   */
  public Vector3D intersection( Vector3D start,
                                Vector3D direction )
  {
    double t1 = _normal.dot( direction );

    // is the direction parallel to the plane?
    if( ( t1 < 0.0001 ) && ( t1 > -0.0001 ) )
    {
      return null;
    }

    double t2 = start.dot( _normal );

    double t3 = ( _distance - t2 ) / t1;

    // Point of collision of line and the plane.
    Vector3D p = start.add( direction.mul( t3 ) );

    Vector3D t4 = p.sub( start );

    double t5 = t4.dot( direction );

    if( t5 > 0 )
    {
      return p;
    }
    else
    {
      return null;
    }
  }

  /**
   * Tests if the specified point lies on this plane.
   * @param point A point in 3-dimentional space.
   * @param maxDistance The maximum allowable distance of the point from the plane.
   * @return true if the point lies on this plane.
   */
  public boolean contains( Vector3D point, double maxDistance )
  {
    double t1 = point.dot( _normal ) - _distance;
    return ( t1 < maxDistance ) && ( t1 > -maxDistance );
  }

  public boolean above( Vector3D point )
  {
    return point.dot( _normal ) > _distance;
  }
}