package liaquay.math.vector;

public class Solid3D
{
  private Vector3D[] _vertices = null;
  private Solid3DFace[] _faces = null;
  private Vector3D[] _normals;
  private double _radius;
  private Vector3D _center;

  private void calculateSolidCenter()
  {
    Vector3D t = new Vector3D(0,0,0);

    for( int i = 0; i < _vertices.length; ++i )
    {
      t = t.add( _vertices[ i ] );
    }
    _center = t.mul( 1.0 / (double)_vertices.length );
  }

  private void calculateSolidRadius()
  {
    double rr = 0.0;

    for( int i = 0; i < _vertices.length; ++i )
    {
      double t = _vertices[ i ].sub( _center ).lengthSquared();
      if( t > rr )
      {
        rr = t;
      }
    }

    _radius = Math.sqrt( rr );
  }

  private void calculateFaceNormals()
  {
    _normals = new Vector3D[ _faces.length ];

    for( int i = 0; i < _faces.length; ++i )
    {
      _normals[ i ] = _faces[ i ].calculateNormal( _vertices );
    }
  }

  public Solid3DFace[] getFaces()
  {
    return _faces;
  }

  public Vector3D[] getVertices()
  {
    return _vertices;
  }

  public Vector3D getFaceNormal( int index )
  {
    return _normals[ index ];
  }

  private Solid3D( Vector3D[] vertices,
                   Solid3DFace[] faces,
                   Vector3D center,
                   double radius,
                   Vector3D[] normals )
  {
    _vertices = vertices;
    _faces = faces;
    _center = center;
    _radius = radius;
  }

  public Solid3D( Vector3D[] vertices, Solid3DFace[] faces)
  {
    _vertices = vertices;
    _faces = faces;

    calculateSolidCenter();
    calculateSolidRadius();
    calculateFaceNormals();
  }

  public final double getRadius()
  {
    return _radius;
  }

  public final double getRadiusSquared()
  {
    return _radius * _radius;
  }

  public Vector3D getCenter()
  {
    return _center;
  }

  public String toString()
  {
    StringBuffer sb = new StringBuffer();
    sb.append( "( " );
    sb.append( "  (\n" );
    for( int i = 0; i < _vertices.length; ++i )
    {
      sb.append( "    (" + _vertices[ i ].get(0) + " " + _vertices[ i ].get(1) + " " + _vertices[ i ].get(2) + ")\n" );
    }
    sb.append( "  )\n" );
    sb.append( "  (\n" );
    for( int i = 0; i < _faces.length; ++i )
    {
      sb.append( "    " + _faces[ i ].toString( _vertices )  + "\n" );
    }
    sb.append( "  )\n" );
    sb.append( ")" );

    return sb.toString();
  }

  public Solid3D mul( Matrix4x3 m )
  {
    Vector3D[] vertices = new Vector3D[ _vertices.length ];

    for( int i = 0; i < _vertices.length; ++i )
    {
      vertices[ i ] = _vertices[ i ].mul( m );
    }

    Vector3D[] normals = new Vector3D[ _faces.length ];

    for( int i = 0; i < _faces.length; ++i )
    {
      normals[ i ] = _normals[ i ].rot( m );
    }

    return new Solid3D( vertices,
                        _faces,
                        _center.mul( m ),
                        _radius,
                        normals );
  }
}