package liaquay.math.vector;

/**
 * Representation of a sphere in 3-dimentional space.
 */
public class Sphere3D
{
  private double _radius;
  private double _radiusSquared;
  private Vector3D _center;

  /**
   * Constructs sphere in 3-dimentional space.
   * @param the radius of the sphere.
   * @param the center of the sphere.
   */
  public Sphere3D( double radius, Vector3D center )
  {
    _radius = radius;
    _radiusSquared = radius * radius;
    _center = center;
  }

  /**
   * Gets the radius of the sphere.
   * @return The radius of the sphere.
   */
  public final double getRadius()
  {
    return _radius;
  }

  /**
   * Gets the center of the sphere.
   * @return The center of the sphere.
   */
  public final Vector3D getCenter()
  {
    return _center;
  }

  /**
   * Sets the radius of the sphere.
   * @param radius The new radius of the sphere.
   */
  public final void setRadius( double radius )
  {
    _radius = radius;
  }

  /**
   * Gets the nearest point of intersection of the sphere and a line.
   * @param line a line in 3-dimentional space.
   * @return The point of intersection of the sphere and the line.
   */
  public Vector3D nearestIntersection( Line3D line )
  {
    return nearestIntersection( line.getPoint(), line.getDirection() );
  }

  /**
   * Gets the nearest point of intersection of the sphere and a line.
   * @param start A point on the line.
   * @param direction The direction of the line.
   * @return The point of intersection of the sphere and the line.
   */
  public Vector3D nearestIntersection( Vector3D start,
                                       Vector3D direction )
  {
    // Vector from the start of the ray to the center of the
    // sphere.
    Vector3D t1 = _center.sub( start );

    // Distance from the start of the ray to a plane
    // perpendicular to the ray, through center of sphere.
    double t2 = t1.dot( direction );

    // Point of collision of line and a plane perpendicular
    // to the ray, through center of sphere.
    Vector3D p = start.add( direction.mul( t2 ) );

    // Distance from the center of the sphere to the
    // collision squared.
    double dd = p.sub( _center ).lengthSquared();

    // Test there is a collision with the sphere.
    if( dd < _radiusSquared )
    {
      // Distance from the center of the sphere to the point
      // of collision of the ray with a plane perpendicular
      // to the ray, through center of sphere.
      double ll = p.sub( _center ).lengthSquared();

      // Distance from the plane to a collision with the
      // surface of the sphere
      double t3 = Math.sqrt( _radiusSquared - ll );

      Vector3D t4 = direction.mul( t3 );
      Vector3D p1 = p.sub( t4 ); // Point of intersection 1
      Vector3D p2 = p.add( t4 ); // Point of intersection 2

      Vector3D t5 = p1.sub( start );
      Vector3D t6 = p2.sub( start );


      double t7 = t5.dot( direction ); // +ve if p1 infront of start
      double t8 = t6.dot( direction ); // +ve if p2 infront of start

      if( t5.lengthSquared() < 0.001 ) t7 = -1;
      if( t6.lengthSquared() < 0.001 ) t8 = -1;

      if( ( t7 > 0.0 ) && ( t8 > 0.0 ) )
      {
        if( t7 < t8 )
        {
          return p1;
        }
        else
        {
          return p2;
        }
      }
      else if( t8 > 0.0 )
      {
        return p2;
      }
      else if( t7 > 0.0 )
      {
        return p1;
      }
      else
      {
        return null; // ??
      }
    }
    else
    {
      return null;
    }
  }

  /**
   * Gets the points of intersection of the sphere and a line.
   * @param start A point on the line.
   * @param direction The direction of the line.
   * @return An array of points of intersection of the sphere and the line.
   */
  public Vector3D[] intersections( Vector3D start,
                                   Vector3D direction )
  {
    // Vector from the start of the ray to the center of the
    // sphere.
    Vector3D t1 = _center.sub( start );

    // Distance from the start of the ray to a plane
    // perpendicular to the ray, through center of sphere.
    double t2 = t1.dot( direction );

    // Point of collision of line and a plane perpendicular
    // to the ray, through center of sphere.
    Vector3D p = start.add( direction.mul( t2 ) );

    // Distance from the center of the sphere to the
    // collision squared.
    double dd = p.sub( _center ).lengthSquared();

    // Test there is a collision with the sphere.
    if( dd < _radiusSquared )
    {
      // Distance from the center of the sphere to the point
      // of collision of the ray with a plane perpendicular
      // to the ray, through center of sphere.
      double ll = p.sub( _center ).lengthSquared();

      // Distance from the plane to a collision with the
      // surface of the sphere
      double t3 = Math.sqrt( _radiusSquared - ll );

      Vector3D t4 = direction.mul( t3 );
      Vector3D p1 = p.sub( t4 ); // Point of intersection 1
      Vector3D p2 = p.add( t4 ); // Point of intersection 2

      return new Vector3D[] { p1, p2 };
    }
    else
    {
      return null;
    }
  }

  public Sphere3D mul( Matrix4x3 m )
  {
    Vector3D center = _center.mul( m );

    return new Sphere3D( _radius, center );
  }
}
