package liaquay.math.vector;

public class Transform
{
  private double _qx;
  private double _qy;
  private double _qz;
  private double _qw;
  private double _dx;
  private double _dy;
  private double _dz;

  public Transform()
  {
    _qx = 0;
    _qy = 0;
    _qz = 0;
    _qw = 1;
    _dx = 0;
    _dy = 0;
    _dz = 0;
  }

  public Transform( Transform t )
  {
    _qx = t._qx;
    _qy = t._qy;
    _qz = t._qz;
    _qw = t._qw;
    _dx = t._dx;
    _dy = t._dy;
    _dz = t._dz;
  }

  public Transform( double qw, double qx, double qy, double qz,
                    double dx, double dy, double dz )
  {
    _qx = qx;
    _qy = qy;
    _qz = qz;
    _qw = qw;
    _dx = dx;
    _dy = dy;
    _dz = dz;
  }

  public final void identity()
  {
    _qx = 0;
    _qy = 0;
    _qz = 0;
    _qw = 1;
    _dx = 0;
    _dy = 0;
    _dz = 0;
  }

  public final void translate( double dx, double dy, double dz)
  {
    _qx = 0;
    _qy = 0;
    _qz = 0;
    _qw = 1;
    _dx = dx;
    _dy = dy;
    _dz = dz;
  }

  public final void rotate( double ux, double uy, double uz, double theta )
  {
    double sinTheta = Math.sin( theta / 2 );
    double cosTheta = Math.cos( theta / 2 );

    _qx = ux * sinTheta;
    _qy = uy * sinTheta;
    _qz = uz * sinTheta;
    _qw = cosTheta;
    _dx = 0;
    _dy = 0;
    _dz = 0;
  }

  public final Transform inverse()
  {
    return inverse( null );
  }

  public final Transform inverse( Transform result )
  {
    double qx = -_qx;
    double qy = -_qy;
    double qz = -_qz;
    double qw =  _qw;
  
    double xx = qx*qx; double yy = qy*qy; double zz = qz*qz;
    double xy = qx*qy; double xz = qx*qz; double yz = qy*qz;
    double wx = qw*qx; double wy = qw*qy; double wz = qw*qz;
    double ww = qw*qw;

    double v00 =   ww + xx - yy - zz;
    double v10 =     2 * ( xy - wz );
    double v20 =     2 * ( xz + wy );

    double v01 =     2 * ( xy + wz );
    double v11 =   ww - xx + yy - zz;
    double v21 =     2 * ( yz - wx );

    double v02 =     2 * ( xz - wy );
    double v12 =     2 * ( yz + wx );
    double v22 =   ww - xx - yy + zz;

    double v0 =
      v00 * _dx +
      v10 * _dy +
      v20 * _dz;

    double v1 =
      v01 * _dx +
      v11 * _dy +
      v21 * _dz;

    double v2 =
      v02 * _dx +
      v12 * _dy +
      v22 * _dz;    
  
    double dx = -v0;
    double dy = -v1;
    double dz = -v2;
  
  
    if( result != null )
    {
      result._qx = qx;
      result._qy = qy;
      result._qz = qz;
      result._qw = qw;
      result._dx = dx;
      result._dy = dy;
      result._dz = dz;
    
      return result;
    }
    else
    {
      return new Transform( qw, qx, qy, qz, dx, dy, dz );
    }
  }

  public final Transform scale( double s )
  {
    double qx = _qx;
    double qy = _qy;
    double qz = _qz;
    double qw = _qw;

    if( ( _qw < 1 ) && ( _qw > -1 ) )
    {
      double theta = Math.acos( _qw );
      double oldSinTheta = Math.sin( theta );

      double ux = _qx / oldSinTheta;
      double uy = _qy / oldSinTheta;
      double uz = _qz / oldSinTheta;

      theta *= s;

      double sinTheta = Math.sin( theta );
      double cosTheta = Math.cos( theta );

      qx = ux * sinTheta;
      qy = uy * sinTheta;
      qz = uz * sinTheta;
      qw = cosTheta;
    }
    else
    {
      qw = 1;
      qx = 0;
      qy = 0;
      qz = 0;
    }

    double dx = _dx * s;
    double dy = _dy * s;
    double dz = _dz * s;

    return new Transform( qw, qx, qy, qz, dx, dy, dz );
  }

  public final void rotateTranslate( double a, double b, double c, double x, double y, double z )
  {
    double sa = Math.sin( a / 2 );
    double ca = Math.cos( a / 2 );
    double sb = Math.sin( b / 2 );
    double cb = Math.cos( b / 2 );
    double sc = Math.sin( c / 2 );
    double cc = Math.cos( c / 2 );


    _qx = (sa * cb * cc) - (ca * sb * sc);
    _qy = (ca * sb * cc) + (sa * cb * sc);
    _qz = (ca * cb * sc) - (sa * sb * cc);
    _qw = (ca * cb * cc) + (sa * sb * sc);

    _dx = x;
    _dy = y;
    _dz = z;
  }

  public final void rotateTranslate( double ux, double uy, double uz, double theta,
                                     double dx, double dy, double dz)
  {
    double sinTheta = Math.sin( theta / 2 );
    double cosTheta = Math.cos( theta / 2 );

    _qx = ux * sinTheta;
    _qy = uy * sinTheta;
    _qz = uz * sinTheta;
    _qw = cosTheta;
    _dx = dx;
    _dy = dy;
    _dz = dz;
  }

  /**
   * Premultiply by t
   */
  public Transform mul( Transform t )
  {
    return mul( t, null );
  }
  
  /**
   * Premultiply by t
   */
  public Transform mul( Transform t, Transform result )
  {
    double qw = ( t._qw * _qw ) - ( t._qx * _qx ) - ( t._qy * _qy ) - ( t._qz * _qz );
    double qx = ( t._qw * _qx ) + ( t._qx * _qw ) + ( t._qy * _qz ) - ( t._qz * _qy );
    double qy = ( t._qw * _qy ) - ( t._qx * _qz ) + ( t._qy * _qw ) + ( t._qz * _qx );
    double qz = ( t._qw * _qz ) + ( t._qx * _qy ) - ( t._qy * _qx ) + ( t._qz * _qw );

    double w = t._qw;
    double x = t._qx;
    double y = t._qy;
    double z = t._qz;

    double xx = x*x; double yy = y*y; double zz = z*z; double ww = w*w;
    double xy = x*y; double xz = x*z; double yz = y*z;
    double wx = w*x; double wy = w*y; double wz = w*z;

    double v00 =   ww + xx - yy - zz;
    double v10 =     2 * ( xy - wz );
    double v20 =     2 * ( xz + wy );

    double v01 =     2 * ( xy + wz );
    double v11 =   ww - xx + yy - zz;
    double v21 =     2 * ( yz - wx );

    double v02 =     2 * ( xz - wy );
    double v12 =     2 * ( yz + wx );
    double v22 =   ww - xx - yy + zz;

    double v0 =
      v00 * _dx +
      v10 * _dy +
      v20 * _dz;

    double v1 =
      v01 * _dx +
      v11 * _dy +
      v21 * _dz;

    double v2 =
      v02 * _dx +
      v12 * _dy +
      v22 * _dz;

    double dx = t._dx + v0;
    double dy = t._dy + v1;
    double dz = t._dz + v2;

    if( result != null )
    {
      result._qx = qx;
      result._qy = qy;
      result._qz = qz;
      result._qw = qw;
      result._dx = dx;
      result._dy = dy;
      result._dz = dz;      
      
      return result;
    }
    else
    {
      return new Transform( qw, qx, qy, qz, dx, dy, dz );
    }
  }

  public final void mop()
  {
    double qx = _qx;
    double qy = _qy;
    double qz = _qz;
    double qw = _qw;

    double length = Math.sqrt( (qx*qx) +  (qy*qy) + (qz*qz) + (qw*qw) );
    qw = qw / length;
    qx = qx / length;
    qy = qy / length;
    qz = qz / length;

    if( ( _qw < 1 ) && ( _qw > -1 ) )
    {
      double theta = Math.acos( _qw );
      double oldSinTheta = Math.sin( theta );

      double ux = _qx / oldSinTheta;
      double uy = _qy / oldSinTheta;
      double uz = _qz / oldSinTheta;

      double l = Math.sqrt( (ux*ux) + (uy*uy) + (uz*uz) );

      ux = ux / l;
      uy = uy / l;
      uz = uz / l;

      double sinTheta = Math.sin( theta );
      double cosTheta = Math.cos( theta );

      _qx = ux * sinTheta;
      _qy = uy * sinTheta;
      _qz = uz * sinTheta;
      _qw = cosTheta;
    }
    else
    {
      _qw = 1;
      _qx = 0;
      _qy = 0;
      _qz = 0;
    }
  }


  /**
   * Premultiply by t
   */
  public Transform inertiaMul( Transform t )
  {
    double qw = ( t._qw * _qw ) - ( t._qx * _qx ) - ( t._qy * _qy ) - ( t._qz * _qz );
    double qx = ( t._qw * _qx ) + ( t._qx * _qw ) + ( t._qy * _qz ) - ( t._qz * _qy );
    double qy = ( t._qw * _qy ) - ( t._qx * _qz ) + ( t._qy * _qw ) + ( t._qz * _qx );
    double qz = ( t._qw * _qz ) + ( t._qx * _qy ) - ( t._qy * _qx ) + ( t._qz * _qw );

    double w = t._qw;
    double x = t._qx;
    double y = t._qy;
    double z = t._qz;

    double dx = t._dx + _dx;
    double dy = t._dy + _dy;
    double dz = t._dz + _dz;

    return new Transform( qw, qx, qy, qz, dx, dy, dz );
  }


  public final double getQx()
  {
    return _qx;
  }

  public final double getQy()
  {
    return _qy;
  }

  public final double getQz()
  {
    return _qz;
  }

  public final double getQw()
  {
    return _qw;
  }

  public final double getDx()
  {
    return _dx;
  }

  public final double getDy()
  {
    return _dy;
  }

  public final double getDz()
  {
    return _dz;
  }

  public final void setDx( double x )
  {
    _dx = x;
  }

  public final void setDy( double y )
  {
    _dy = y;
  }

  public final void setDz( double z )
  {
    _dz = z;
  }

  public final String toString()
  {
    double ux;
    double uy;
    double uz;
    double theta;
    double l;

    if( ( _qw < 1 ) && ( _qw > -1 ) )
    {
      theta = Math.acos( _qw ) * 2;
      double sinTheta = Math.sin( theta / 2 );

      ux = _qx / sinTheta;
      uy = _qy / sinTheta;
      uz = _qz / sinTheta;

      l = Math.sqrt( (ux*ux) + (uy*uy) + (uz*uz) );

      ux = ux / l;
      uy = uy / l;
      uz = uz / l;
    }
    else
    {
      theta = 0;
      ux = 0;
      uy = 0;
      uz = 0;
      l = 1;
    }

    String s1 = "Angle "+theta+" Unit vector ("+ux+", "+uy+", "+uz+" ) length=" + l;
    String s2 = "Quaternion w="+_qw + " x=" + _qx + " y=" + _qy + " z=" + _qz ;
    String s3 = "Offset (" + _dx + ", " + _dy + ", " + _dz +" )";
    return s1 + "\n" + s2 + "\n" + s3;
  }
  
  public static Transform createIdentity()
  {
    return new Transform();
  }

  public static Transform createRotateTranslate( double a, double b, double c, double x, double y, double z )
  {
    Transform m = new Transform();
    m.rotateTranslate( a,b,c,x,y,z );
    return m;
  }  
}