package liaquay.math.vector;

/**
 * Representation of a 2-dimentional vector.
 */
public class Vector2D
{
  /**
   * The components of the vector as an array [x, y].
   */
  private double[] _values;

  /**
   * Constructs a vector given the x,y and z components.
   * @param x The x component.
   * @param y The y component.
   * @param z The z component.
   */
  public Vector2D( double x,
                   double y )
  {
    _values = new double[ 2 ];

    _values[ 0 ] = x;
    _values[ 1 ] = y;
  }

  /**
   * Constructs a vector given the x and y components in an array.
   * @param values An array [x,y].
   */
  public Vector2D( double[] values )
  {
    _values = values;
  }

  /**
   * Get the components of the vector as an array [x, y].
   * @return The components of the vector as an array [x, y].
   */
  public final double[] getValues()
  {
    return _values;
  }

  /**
   * Add a vector to this vector and return the result.
   * <p> This vector is v1
   * <p> v3 = v1 + v2
   * <p> [x3, y3] = [x1+x2, y1+y2]
   * <p>
   * @param v2 The vector to add to this vector.
   * @return The sum of this vector and v2
   */
  public final Vector2D add( Vector2D v2 )
  {
    double[] r = v2.getValues();
    double[] v = new double[ 3 ];

    v[ 0 ] = _values[ 0 ] + r [ 0 ];
    v[ 1 ] = _values[ 1 ] + r [ 1 ];

    return new Vector2D( v );
  }

  /**
   * Multiplies this vector by a scaler value and returns the result.
   * <p> This vector is v1
   * <p> v2 = v1 * s
   * <p> [x2, y2] = [x1*s, y1*s]
   * <p>
   * @param s The scaler by which to multiply this vector.
   * @return This vector multiplied by the scaler s.
   */
  public final Vector2D mul( double s )
  {
    double[] v = new double[ 2 ];

    v[ 0 ] = _values[ 0 ] * s;
    v[ 1 ] = _values[ 1 ] * s;

    return new Vector2D( v );
  }

  /**
   * Returns the length of this vector.
   * <p> length = sqrt( (x*x) + (y*y) )
   * <p>
   * @return The length of this vector.
   */
  public final double length()
  {
    return Math.sqrt( lengthSquared() );
  }

  /**
   * Returns the length of this vector squared.
   * <p> lengthSquared = (x*x) + (y*y)
   * <p>
   * @return The length of this vector squared.
   */
  public final double lengthSquared()
  {
    double v[] = _values;
    return ( ( v[ 0 ] * v [ 0 ] ) + ( v[ 1 ] * v [ 1 ] ) );
  }

  /**
   * Scales this vector such that its length is 1.
   * <p> length = sqrt( (x*x) + (y*y) )
   * <p> [x2, y2, z2] = [ x1/length, y1/length, z1/length ]
   * <p>
   * @return Unit vector in the direction of this vector.
   */
  public final Vector2D normalize()
  {
    double l = length();

    return new Vector2D( _values[ 0 ] / l, _values[ 1 ] / l );
  }

  /**
   * Dot a vector with this vector and return the result.
   * <p> This vector is v1
   * <p> s1 = v1.v2
   * <p> s1 = (x1*x2) + (y1*y2)
   * <p>
   * @param m The vector to dot with this vector.
   * @return The dot product of this vector and v2
   */
  public final double dot( Vector2D v2 )
  {
    double v[] = v2.getValues();

    return ( _values[ 0 ] * v[ 0 ] ) + ( _values[ 1 ] * v[ 1 ] );
  }

  /**
   * Subtract a vector from this vector and return the result.
   * <p> This vector is v1
   * <p> v3 = v1 - v2
   * <p> [x3, y3] = [x1-x2, y1-y2]
   * <p>
   * @param v2 The vector to subtract from this vector.
   * @return This vector minus v2
   */
  public final Vector2D sub( Vector2D v2 )
  {
    double v[] = v2.getValues();
    double newValues[] = new double[ 2 ];

    newValues[ 0 ] = _values[ 0 ] - v[ 0 ];
    newValues[ 1 ] = _values[ 1 ] - v[ 1 ];

    return new Vector2D( newValues );
  }

  /**
   * Get a component of the vector.
   * <p> i = 0,1 component = x,y
   * @param i The index of the component of the vector to return.
   * @return The component of the vector specified by the index.
   */
  public final double get( int i )
  {
    return _values[ i ];
  }

  /**
   * Set a component of the vector.
   * <p> i = 0,1 component = x,y
   * @param i The index of the component of the vector to return.
   * @param f The value of the component of the vector.
   */
  public final void set( int i, double f )
  {
    _values[ i ] = f;
  }

  public String toString()
  {
    return "(" + _values[ 0 ] + " " + _values[ 1 ] + ")";
  }
  
  
  public boolean equals( Object o )
  {
    Vector2D v = (Vector2D)o;
    
    return ( _values[ 0 ] == v._values[ 0 ] ) &&
           ( _values[ 1 ] == v._values[ 1 ] );
  }  
}
