package liaquay.math.vector;

/**
 * Representation of a 3-dimentional vector.
 */
public class Vector3D
{
  /**
   * The components of the vector as an array [x, y, z].
   */
  private double[] _values;

  /**
   * Constructs a vector given the x,y and z components.
   * @param x The x component.
   * @param y The y component.
   * @param z The z component.
   */
  public Vector3D( double x,
                   double y,
                   double z )
  {
    _values = new double[ 3 ];

    _values[ 0 ] = x;
    _values[ 1 ] = y;
    _values[ 2 ] = z;
  }

  /**
   * Constructs a vector given the x,y and z components in an array.
   * @param values An array [x,y,z].
   */
  public Vector3D( double[] values )
  {
    _values = values;
  }

  /**
   * Get the components of the vector as an array [x, y, z].
   * @return The components of the vector as an array [x, y, z].
   */
  public final double[] getValues()
  {
    return _values;
  }

  /**
   * Add a vector to this vector and return the result.
   * <p> This vector is v1
   * <p> v3 = v1 + v2
   * <p> [x3, y3, z3] = [x1+x2, y1+y2, z1+z2]
   * <p>
   * @param v2 The vector to add to this vector.
   * @return The sum of this vector and v2
   */
  public final Vector3D add( Vector3D v2 )
  {
    double[] r = v2.getValues();
    double[] v = new double[ 3 ];

    v[ 0 ] = _values[ 0 ] + r [ 0 ];
    v[ 1 ] = _values[ 1 ] + r [ 1 ];
    v[ 2 ] = _values[ 2 ] + r [ 2 ];

    return new Vector3D( v );
  }

  /**
   * Multiplies this vector by a scaler value and returns the result.
   * <p> This vector is v1
   * <p> v2 = v1 * s
   * <p> [x2, y2, z2] = [x1*s, y1*s, z1*s]
   * <p>
   * @param s The scaler by which to multiply this vector.
   * @return This vector multiplied by the scaler s.
   */
  public final Vector3D mul( double s, Vector3D result )
  {
    double[] v = result.getValues();

    v[ 0 ] = _values[ 0 ] * s;
    v[ 1 ] = _values[ 1 ] * s;
    v[ 2 ] = _values[ 2 ] * s;

    return result;
  }

  /**
   * Multiplies this vector by a scaler value and returns the result.
   * <p> This vector is v1
   * <p> v2 = v1 * s
   * <p> [x2, y2, z2] = [x1*s, y1*s, z1*s]
   * <p>
   * @param s The scaler by which to multiply this vector.
   * @return This vector multiplied by the scaler s.
   */
  public final Vector3D mul( double s )
  {
    double[] v = new double[ 3 ];

    v[ 0 ] = _values[ 0 ] * s;
    v[ 1 ] = _values[ 1 ] * s;
    v[ 2 ] = _values[ 2 ] * s;

    return new Vector3D( v );
  }

  /**
   * Pre-multiplies this vector by the matrix m.
   * <p> This vector is v1
   * <p> v2 = m * v1
   * @param m A 4x3 matrix by which to pre-multiply this vector.
   * @return A Vector3D transformed by the matrix m.
   */
  public final Vector3D mul( Matrix4x3 m )
  {
    return mul( m, null );
  }

  /**
   * Pre-multiplies this vector by the matrix m.
   * <p> This vector is v1
   * <p> result = m * v1
   * @param m A 4x3 matrix by which to pre-multiply this vector.
   * @param result This vector transformed by the matrix m.
   */
  public final Vector3D mul( Matrix4x3 m, Vector3D result )
  {
    double[][] r = m.getValues();
    double[] v;

    if( result == null )
    {
      v = new double[ 3 ];
    }
    else
    {
      v = result.getValues();
    }

    double v0 =
      r[ 0 ][ 0 ] * _values[ 0 ] +
      r[ 1 ][ 0 ] * _values[ 1 ] +
      r[ 2 ][ 0 ] * _values[ 2 ] +
      r[ 3 ][ 0 ] ;

    double v1 =
      r[ 0 ][ 1 ] * _values[ 0 ] +
      r[ 1 ][ 1 ] * _values[ 1 ] +
      r[ 2 ][ 1 ] * _values[ 2 ] +
      r[ 3 ][ 1 ] ;

    double v2 =
      r[ 0 ][ 2 ] * _values[ 0 ] +
      r[ 1 ][ 2 ] * _values[ 1 ] +
      r[ 2 ][ 2 ] * _values[ 2 ] +
      r[ 3 ][ 2 ] ;

    v[ 0 ] = v0;
    v[ 1 ] = v1;
    v[ 2 ] = v2;

    if( result == null )
    {
      return new Vector3D( v );
    }
    else
    {
      return result;
    }
  }

  public final Vector3D rot( Matrix4x3 m )
  {
    return rot( m, null );
  }

  public final Vector3D rot( Matrix4x3 m, Vector3D result )
  {
    double[][] r = m.getValues();
    double[] v;

    if( result == null )
    {
      v = new double[ 3 ];
    }
    else
    {
      v = result.getValues();
    }

    double v0 =
      r[ 0 ][ 0 ] * _values[ 0 ] +
      r[ 1 ][ 0 ] * _values[ 1 ] +
      r[ 2 ][ 0 ] * _values[ 2 ];

    double v1 =
      r[ 0 ][ 1 ] * _values[ 0 ] +
      r[ 1 ][ 1 ] * _values[ 1 ] +
      r[ 2 ][ 1 ] * _values[ 2 ];

    double v2 =
      r[ 0 ][ 2 ] * _values[ 0 ] +
      r[ 1 ][ 2 ] * _values[ 1 ] +
      r[ 2 ][ 2 ] * _values[ 2 ];

    v[ 0 ] = v0;
    v[ 1 ] = v1;
    v[ 2 ] = v2;

    if( result == null )
    {
      return new Vector3D( v );
    }
    else
    {
      return result;
    }
  }

  /**
   * Gets the x component of the perspective transform.
   * <p> perx = ( x * ppd ) / z
   * @param ppd ??
   * @return Perspective transformed x coodinate.
   */
  public final double perx( double ppd )
  {
    return _values[ 0 ] * ppd / _values[ 2 ];
  }

  /**
   * Gets the y component of the perspective transform.
   * <p> perx = ( y * ppd ) / z
   * @param ppd ??
   * @return Perspective transformed y coodinate.
   */
  public final double pery( double ppd )
  {
    return _values[ 1 ] * ppd / _values[ 2 ];
  }

  /**
   * Returns the length of this vector.
   * <p> length = sqrt( (x*x) + (y*y) + (z*z) )
   * <p>
   * @return The length of this vector.
   */
  public final double length()
  {
    return Math.sqrt( lengthSquared() );
  }

  /**
   * Returns the length of this vector squared.
   * <p> lengthSquared = (x*x) + (y*y) + (z*z)
   * <p>
   * @return The length of this vector squared.
   */
  public final double lengthSquared()
  {
    double v[] = _values;
    return ( ( v[ 0 ] * v [ 0 ] ) + ( v[ 1 ] * v [ 1 ] ) + ( v[ 2 ] * v [ 2 ] ) );
  }

  /**
   * Scales this vector such that its length is 1.
   * <p> length = sqrt( (x*x) + (y*y) + (z*z) )
   * <p> [x2, y2, z2] = [ x1/length, y1/length, z1/length ]
   * <p>
   * @return Unit vector in the direction of this vector.
   */
  public final Vector3D normalize()
  {
    double l = length();

    return new Vector3D( _values[ 0 ] / l, _values[ 1 ] / l, _values[ 2 ] / l );
  }

  /**
   * Dot a vector with this vector and return the result.
   * <p> This vector is v1
   * <p> s1 = v1.v2
   * <p> s1 = (x1*x2) + (y1*y2) + (z1*z2)
   * <p>
   * @param m The vector to dot with this vector.
   * @return The dot product of this vector and v2
   */
  public final double dot( Vector3D v2 )
  {
    double v[] = v2.getValues();

    return ( _values[ 0 ] * v[ 0 ] ) + ( _values[ 1 ] * v[ 1 ] ) + ( _values[ 2 ] * v[ 2 ] );
  }

  /**
   * Subtract a vector from this vector and return the result.
   * <p> This vector is v1
   * <p> v3 = v1 - v2
   * <p> [x3, y3, z3] = [x1-x2, y1-y2, z1-z2]
   * <p>
   * @param v2 The vector to subtract from this vector.
   * @return This vector minus v2
   */
  public final Vector3D sub( Vector3D v2 )
  {
    double v[] = v2.getValues();
    double newValues[] = new double[ 3 ];

    newValues[ 0 ] = _values[ 0 ] - v[ 0 ];
    newValues[ 1 ] = _values[ 1 ] - v[ 1 ];
    newValues[ 2 ] = _values[ 2 ] - v[ 2 ];

    return new Vector3D( newValues );
  }

  /**
   * Gets the cross product of this vector and v2.
   * <p> This vector is v1
   * <p> v3 = v1 X v2
   * <p> [x3, y3, z3] = [ (y2*z1)-(y1*z2), (z2*x1)-(z1*x2),(x2*1y)-(x1*y2)]
   * <p>
   * @return The cross product of this vector and v2.
   */
  public final Vector3D cross( Vector3D v2 )
  {
    double v[] = v2.getValues();
    double newValues[] = new double[ 3 ];

    newValues[ 0 ] = ( v[ 1 ] * _values[ 2 ] ) - ( _values[ 1 ] * v[ 2 ] );
    newValues[ 1 ] = ( v[ 2 ] * _values[ 0 ] ) - ( _values[ 2 ] * v[ 0 ] );
    newValues[ 2 ] = ( v[ 0 ] * _values[ 1 ] ) - ( _values[ 0 ] * v[ 1 ] );

    return new Vector3D( newValues );
  }

  /**
   * Get a component of the vector.
   * <p> i = 0,1,2 component = x,y,z
   * @param i The index of the component of the vector to return.
   * @return The component of the vector specified by the index.
   */
  public final double get( int i )
  {
    return _values[ i ];
  }

  /**
   * Set a component of the vector.
   * <p> i = 0,1,2 component = x,y,z
   * @param i The index of the component of the vector to return.
   * @param f The value of the component of the vector.
   */
  public final void set( int i, double f )
  {
    _values[ i ] = f;
  }

  public final Vector3D mul( Transform m )
  {
    return mul( m, null );
  }

  public final Vector3D mul( Transform t, Vector3D result )
  {
    double[] v;

    if( result == null )
    {
      v = new double[ 3 ];
    }
    else
    {
      v = result.getValues();
    }

    double x = t.getQx();
    double y = t.getQy();
    double z = t.getQz();
    double w = t.getQw();

    double xx = x*x; double yy = y*y; double zz = z*z;
    double xy = x*y; double xz = x*z; double yz = y*z;
    double wx = w*x; double wy = w*y; double wz = w*z;
    double ww = w*w;

    double v00 =   ww + xx - yy - zz;
    double v10 =     2 * ( xy - wz );
    double v20 =     2 * ( xz + wy );

    double v01 =     2 * ( xy + wz );
    double v11 =   ww - xx + yy - zz;
    double v21 =     2 * ( yz - wx );

    double v02 =     2 * ( xz - wy );
    double v12 =     2 * ( yz + wx );
    double v22 =   ww - xx - yy + zz;

    double v0 =
      v00 * _values[ 0 ] +
      v10 * _values[ 1 ] +
      v20 * _values[ 2 ];

    double v1 =
      v01 * _values[ 0 ] +
      v11 * _values[ 1 ] +
      v21 * _values[ 2 ];

    double v2 =
      v02 * _values[ 0 ] +
      v12 * _values[ 1 ] +
      v22 * _values[ 2 ];

    v[ 0 ] = v0 + t.getDx();
    v[ 1 ] = v1 + t.getDy();
    v[ 2 ] = v2 + t.getDz();

    if( result == null )
    {
      return new Vector3D( v );
    }
    else
    {
      return result;
    }
  }

  public final Vector3D rot( Transform m )
  {
    return rot( m, null );
  }

  public final Vector3D rot( Transform t, Vector3D result )
  {
    double[] v;

    if( result == null )
    {
      v = new double[ 3 ];
    }
    else
    {
      v = result.getValues();
    }

    double x = t.getQx();
    double y = t.getQy();
    double z = t.getQz();
    double w = t.getQw();

    double xx = x*x; double yy = y*y; double zz = z*z;
    double xy = x*y; double xz = x*z; double yz = y*z;
    double wx = w*x; double wy = w*y; double wz = w*z;
    double ww = w*w;

    double v00 =   ww + xx - yy - zz;
    double v10 =     2 * ( xy - wz );
    double v20 =     2 * ( xz + wy );

    double v01 =     2 * ( xy + wz );
    double v11 =   ww - xx + yy - zz;
    double v21 =     2 * ( yz - wx );

    double v02 =     2 * ( xz - wy );
    double v12 =     2 * ( yz + wx );
    double v22 =   ww - xx - yy + zz;

    double v0 =
      v00 * _values[ 0 ] +
      v10 * _values[ 1 ] +
      v20 * _values[ 2 ];

    double v1 =
      v01 * _values[ 0 ] +
      v11 * _values[ 1 ] +
      v21 * _values[ 2 ];

    double v2 =
      v02 * _values[ 0 ] +
      v12 * _values[ 1 ] +
      v22 * _values[ 2 ];

    v[ 0 ] = v0;
    v[ 1 ] = v1;
    v[ 2 ] = v2;

    if( result == null )
    {
      return new Vector3D( v );
    }
    else
    {
      return result;
    }
  }

  public final Vector3D invert()
  {
    return new Vector3D( -_values[ 0 ], -_values[ 1 ], -_values[ 2 ] );
  }

  public String toString()
  {
    return "(" + _values[ 0 ] + " " + _values[ 1 ] + " " + _values[ 2 ] + ")";
  }

  public final boolean equals( Object o )
  {
    Vector3D v = (Vector3D)o;

    return ( _values[ 0 ] == v._values[ 0 ] ) &&
           ( _values[ 1 ] == v._values[ 1 ] ) &&
           ( _values[ 2 ] == v._values[ 2 ] );
  }
}
